#!/usr/bin/env python3

##---------------------------Imports-----------------------------------------------##
import os, sys
import argparse
import re


CURR_DIR = os.path.dirname (os.path.abspath (__file__) )
BASE_DIR = os.path.dirname (os.path.dirname (os.path.dirname (CURR_DIR) ) )
sys.path.insert (0, os.path.join (BASE_DIR, 'Tools') )
sys.path.insert (0, CURR_DIR)


from Arguments import Arguments



##---------------------------Start Main--------------------------------------------##
if __name__ != "__main__":
   print ("error: must be called as program")
   exit (1)


##---------------------------Read Arguments----------------------------------------##
parser = argparse.ArgumentParser ()
parser.add_argument ('--config', help='Configuration file', type=str, default='')
parser.add_argument ('--verbose', help='Verbose', type=int, default=2)
(args, remaining) = parser.parse_known_args ()


##---------------------------Create midlayer database------------------------------##
import converters

db = converters.database.load (args.config, args.verbose)

import console; console()


##---------------------------Start vprint------------------------------------------##
def vprint (v, *argv, **kwargs):
   if args.verbose >= v:
      print (*argv, **kwargs)


##---------------------------Migrate into Neo4j database---------------------------##
from Neo4j import Neo4j

neo4j = Neo4j ()


##---------------------------Migrate nodes-----------------------------------------##
stats = {'AF': 0, 'AT': 0, 'CT': 0, 'GO': 0, 'KEGGX': 0, 'OB': 0, 'PW': 0, 'SP': 0, 'edges': 0}

try:
   # # AgeingFactor
   # vprint (2, "Create Nodes of type AgeingFactor"); i=0
   # for AF_id, af in db.AF:
   #    if i % 1000 == 0: vprint (2, "Node %d/%d" % (i, len (db.AF) ), end = '\r', flush = True)
   #    af.AF_name = af.AF_name.replace ("'", "\\'")
   #    neo4j.call ('nodes/AgeingFactor', af)
   #    for sp in af.AF_species:
   #       sp = re.sub ('(?![a-zA-Z]).', '_', sp)
   #       neo4j.call ('commands/setAF_Species', {'AF_id': af.AF_id, 'SP_label': sp})
   #    stats['AF'] += 1; i+=1

   # # AlleleType
   # vprint (2, "Create Nodes of type AlleleType"); i=0
   # for AT_id, at in db.AT:
   #    if i % 1000 == 0: vprint (2, "Node %d/%d" % (i, len (db.AT) ), end = '\r', flush = True)
   #    neo4j.call ('nodes/AlleleType', at)
   #    stats['AT'] += 1; i+=1

   # # Citation
   # vprint (2, "Create Nodes of type Citation"); i=0
   # for CT_id, ct in db.CT:
   #    if i % 1000 == 0: vprint (2, "Node %d/%d" % (i, len (db.CT) ), end = '\r', flush = True)
   #    neo4j.call ('nodes/Citation', ct)
   #    stats['CT'] += 1; i+=1

   # # GeneOntologyTerm
   # vprint (2, "Create Nodes of type GeneOntologyTerm"); i=0
   # for GO_id, go in db.GO:
   #    if i % 1000 == 0: vprint (2, "Node %d/%d" % (i, len (db.GO) ), end = '\r', flush = True)
   #    go.GO_term = go.GO_term.replace ("'", "\\'")
   #    neo4j.call ('nodes/GeneOntologyTerm', go)
   #    stats['GO'] += 1; i+=1

   # # KEGGPathwayExtraGenes
   # vprint (2, "Create Nodes of type KEGGPathwayExtraGenes"); i=0
   # for KEGGX_id, keggx in db.KEGGX:
   #    if i % 1000 == 0: vprint (2, "Node %d/%d" % (i, len (db.KEGGX) ), end = '\r', flush = True)
   #    keggx.KEGGX_description = keggx.KEGGX_description.replace ("'", "\\'")
   #    keggx.KEGGX_symbol      = keggx.KEGGX_symbol.replace ("'", "\\'")
   #    neo4j.call ('nodes/KEGGPathwayExtraGenes', keggx)
   #    stats['KEGGX'] += 1; i+=1

   # # # Observation
   # vprint (2, "Create Nodes of type Observation"); i=0
   # for OB_id, ob in db.OB:
   #    if i % 1000 == 0: vprint (2, "Node %d/%d" % (i, len (db.OB) ), end = '\r', flush = True)
   #    if 'OB_lifespan_effect_significant' not in ob: ob.OB_lifespan_effect_significant = 'null'
   #    else: ob.OB_lifespan_effect_significant = "'" + ob.OB_lifespan_effect_significant + "'"
   #    if 'OB_lifespan_change' not in ob: ob.OB_lifespan_change = 'null'
   #    else: ob.OB_lifespan_change = "'" + ob.OB_lifespan_change + "'"
   #    if 'OB_lifespan_measure' not in ob: ob.OB_lifespan_measure = 'null'
   #    else: ob.OB_lifespan_measure = "'" + ob.OB_lifespan_measure + "'"
   #    if 'OB_homology_database' not in ob: ob.OB_homology_database = 'null'
   #    else: ob.OB_homology_database = "'" + ob.OB_homology_database + "'"
   #    if 'OB_homology_group_id' not in ob: ob.OB_homology_group_id = 'null'
   #    else: ob.OB_homology_group_id = "'" + ob.OB_homology_group_id + "'"

   #    neo4j.call ('nodes/Observation', ob)

   #    for sp in ob.OB_species:
   #       sp = re.sub ('(?![a-zA-Z]).', '_', sp)
   #       neo4j.call ('commands/setOB_Species', {'OB_id': ob.OB_id, 'SP_label': sp})

   #    stats['OB'] += 1; i+=1

   # Pathway
   vprint (2, "Create Nodes of type Pathway"); i=0
   for PW_id, pw in db.PW:
      if i % 1000 == 0: vprint (2, "Node %d/%d" % (i, len (db.PW) ), end = '\r', flush = True)
      if pw.PW_description: pw.PW_description = pw.PW_description.replace ("'", "\\'")
      if not pw.PW_description: pw.PW_description = 'null'
      else: pw.PW_description = "'" + pw.PW_description + "'"
      if not pw.PW_ncbi_tax_id: pw.PW_ncbi_tax_id = 'null'
      else: pw.PW_ncbi_tax_id = "'" + pw.PW_ncbi_tax_id + "'"
      pw.PW_name        = pw.PW_name.replace ("'", "\\'")
      neo4j.call ('nodes/Pathway', pw)
      stats['PW'] += 1; i+=1

   # # Species
   # vprint (2, "Create Nodes of type Species"); i=0
   # for SP_id, sp in db.SP:
   #    if i % 1000 == 0: vprint (2, "Node %d/%d" % (i, len (db.SP) ), end = '\r', flush = True)
   #    neo4j.call ('nodes/Species', sp)
   #    stats['SP'] += 1; i+=1

except Exception as ex:
   print ("Exception: " + str (ex), file = sys.stderr)
   import console; console()

#import console; console()


##---------------------------Migrate edges-----------------------------------------##
# try:
#    mapping = {
#       'AF': 'AgeingFactor',
#       'AT': 'AlleleType',
#       'CT': 'Citation',
#       'GO': 'GeneOntologyTerm',
#       'KEGGX': 'KEGGPathwayExtraGenes',
#       'OB': 'Observation',
#       'PW': 'Pathway',
#       'SP': 'Species'
#    }

#    # Example:
#    # type = OB-AL-AF
#    # OB -> get/Observation, AL -> , AF -> get/AgeingFactor
#    # Final Cypher call (concatenated)
#    #   get/Observation, get/AgeingFactor, links/OB-AL-AF

#    vprint (2, "Create Edges")
#    for i, edge in enumerate (db.edges):
#       if i % 1000 == 0: vprint (2, "Edge %d/%d" % (i, len (db.edges) ), end = '\r', flush = True)

#       nodes = edge['type'].split ('-')
#       files = []

#       for node in nodes:
#          if node == 'AL' or node == 'L': continue
#          files.append ('get/' + mapping[node])

#       files.append ('links/' + edge['type'])
#       edge.update (db.edgesLabels)

#       neo4j.call (files, edge)
#       stats['edges'] += 1

# except Exception as ex:
#    print ("Exception: " + str (ex), file = sys.stderr)
#    import console; console()


vprint (2, "Import succedded")

# import console; console()
