
##---------------------------Imports-----------------------------------------------##
import os, sys

import check

from Arguments import Arguments



##---------------------------Start load--------------------------------------------##
def load (filename, db):
   with open (filename, 'r') as file:
      header = file.readline ().strip ().split ('\t')

      for i, line in enumerate (file):
         tline = line.strip ()
         if len (tline) == 0: continue

         temp1 = tline.split ('\t')
         temp2 = Arguments ()

         # Glue header with values
         for h, t in zip (header, temp1): temp2[h] = t


         temp2.entry_stable_id = 'AF_%06d' % int (temp2.entry_id)

         # Check consistency
         if temp2.entry_stable_id not in db.AF: raise Exception ("Missing Node '%s'" % temp2.entry_stable_id)

         af = db.AF[temp2.entry_stable_id]

         check.setValue (af, 'AF_type', temp2.entry_type)
         check.setValue (af, 'AF_name', temp2.entry_name)


         # Observation
         if temp2.observation_stable_id not in db.OB:
            species = [temp2.species] if temp2.species else []
            AT_id = db.ATmapping[temp2.allele_types] if temp2.allele_types in db.ATmapping else temp2.allele_types

            db.OB[temp2.observation_stable_id] = Arguments (
               OB_id                          = temp2.observation_stable_id,
               OB_lifespan_effect_significant = temp2.lifespan_effect_significant,
               OB_lifespan_change             = temp2.lifespan_change,
               OB_lifespan_measure            = temp2.lifespan_measure,
               #OB_homology_database           = 'HomoloGene',   # Hack, is normally dependent of homologene_id
               #OB_homology_group_id           = temp2.homologene_id,

               # Edges
               OB_species                     = set (species),
               OB_ageing_factors              = set ([temp2.entry_stable_id]),
               OB_base_ageing_factors         = set (),
               OB_allele_types                = [{'AF_id': temp2.entry_stable_id, 'AT_id': AT_id, 'AL_type': temp2.allele_types, 'AL_base': temp2.allele_types_base}]
            )

         else:
            ob = db.OB[temp2.observation_stable_id]

            check.setValue (ob, 'OB_lifespan_effect_significant', temp2.lifespan_effect_significant)
            check.setValue (ob, 'OB_lifespan_change', temp2.lifespan_change)
            check.setValue (ob, 'OB_lifespan_measure', temp2.lifespan_measure)

            if temp2.species: ob.OB_species.add (temp2.species)
            ob.OB_ageing_factors.add (temp2.entry_stable_id)
            AT_id = db.ATmapping[temp2.allele_types] if temp2.allele_types in db.ATmapping else temp2.allele_types
            ob.OB_allele_types.append ({'AF_id': temp2.entry_stable_id, 'AT_id': AT_id, 'AL_type': temp2.allele_types, 'AL_base': temp2.allele_types_base})


         # AF-Species
         if temp2.species: db.AF[temp2.entry_stable_id].AF_species.add (temp2.species)


         # AlleleType
         AT_id = db.ATmapping[temp2.allele_types] if temp2.allele_types in db.ATmapping else temp2.allele_types
         if AT_id and AT_id not in db.AT:
            db.AT[AT_id] = Arguments (
               AT_id = AT_id
            )


         # Species
         if temp2.species and temp2.species not in db.SP:
            db.SP[temp2.species] = Arguments (
               SP_id = temp2.species
            )
