#!/usr/bin/env python3

import sys, os
import re

from neo4j.v1 import GraphDatabase


CURR_DIR = os.path.dirname (os.path.abspath (__file__) )
BASE_DIR = os.path.dirname (CURR_DIR)
sys.path.insert (0, CURR_DIR)

from Arguments import Arguments



##---------------------------Class Neo4j-------------------------------------------##
class Neo4j:
   ##------------------------Start Constructor-------------------------------------##
   def __init__ (self, settings = os.path.join (CURR_DIR, 'settings.ini') ):
      config = Arguments ()
      config.load (settings)
      self.db = GraphDatabase.driver (config.url, auth = config["user", "password"])



   ##------------------------Start createSession-----------------------------------##
   def createSession (self):
      return self.db.session()



   ##------------------------Start call--------------------------------------------##
   def call (self, filenames, config, session = None, **kwargs):
      cypher = self._parseCyphers (filenames, config)
      return self.eval (cypher, session = session, **kwargs)



   ##------------------------Start call--------------------------------------------##
   def eval (self, cypher, session = None, **kwargs):
      if session:
         return session.run (cypher, **kwargs)
      else:
         with self.db.session() as session:
            return session.run (cypher, **kwargs)



   ##------------------------Start _parseCypher------------------------------------##
   def _parseCypher (self, filename, config):
      if not os.path.isfile (filename):
         filename = os.path.join (BASE_DIR, "Cypher", filename + '.cypher')
      if not os.path.isfile (filename):
         raise Exception ("File not found '%s'" % (filename) )

      with open (filename, 'r') as file:
         cypher = file.read ()
         #for line in file:
         #   if line.lstrip()[0] != '#':
         #      cypher += line

         def replace (m):
            key = m.group(1)
            value = config[key]
            if type (value) == bool:
               return ('true' if value else 'false')
            elif type (value) == list:
               return str (value)[1:-1]
            elif type (value) == set:
               return str (list (value) )[1:-1]
            else:
               return str (value)

         return re.sub ('\$\{([a-zA-Z0-9_-]+)\}', replace, cypher)



   ##------------------------Start _parseCyphers-----------------------------------##
   def _parseCyphers (self, filenames, config):
      if type (filenames)  == str:
         filenames = [filenames]

      res = ""

      for filename in filenames:
         res += self._parseCypher (filename, config)
         if res[-1] != '\n': res += '\n'

      return res
