
##---------------------------Inlcudes----------------------------------------------##



__all__ = ["shell"]

##------------------------Start  _readline-----------------------------------------##
#TODO
def _readline (_locals, _globals):
   # https://gist.github.com/twneale/5245670

   if 'write_history_file' not in _globals:
      import os
      import readline        # optional, will allow Up/Down/History in the console
      import rlcompleter
      import atexit

      python_path  = os.path.join (os.environ['HOME'], '.config/python')
      history_file = os.path.join (python_path, 'history')

      try:
         readline.read_history_file (history_file)
      except IOError:
         pass

      readline.parse_and_bind ("tab: complete")
      readline.set_history_length (1000)

      def write_history_file (python_path, history_file):
         import os, readline
         if not os.path.isdir (python_path):
            os.makedirs (python_path)
         readline.write_history_file (history_file)

      atexit.register (write_history_file, python_path, history_file)

      _globals.update ({'write_history_file': write_history_file})

      del os, readline, rlcompleter, atexit, python_path, history_file

   if 'bt' not in _globals:
      def bt ():
         # https://stackoverflow.com/questions/2359248/how-can-you-programmatically-inspect-the-stack-trace-of-an-exception-in-python
         import inspect
         for stk in inspect.stack ()[1:][::-1]:
            print ("  File \"%s\", line %d, in %s" % (stk.filename, stk.lineno, stk.function) )


      _globals.update ({'bt': bt})



##------------------------Start Shell----------------------------------------------##
def shell (_locals = None, _globals = None):
   """
      Switch into an interactive shell for more advanced debugging
   """

   import sys, code

   # https://stackoverflow.com/questions/6618795/get-locals-from-calling-namespace-in-python
   if not _locals or not _globals:
      import inspect
      if not _locals:  _locals  = inspect.currentframe ().f_back.f_locals
      if not _globals: _globals = inspect.currentframe ().f_back.f_globals


   def namespace ():
      namespace = {}
      namespace.update (_globals)
      namespace.update (_locals)
      return namespace


   if 'IPython.terminal' in sys.modules:      # IPython
      from IPython import embed

      #embed (local_ns = _locals, global_ns = _globals)   # Start interactive shell
      #embed (user_ns = _locals, user_module = globals)   # Start interactive shell
      embed (user_ns = namespace () )   # Start interactive shell

      #from IPython.terminal import embed
      #shell = embed.InteractiveShellEmbed.instance ()
      #shell (stack_depth = 2, local_ns = _locals, global_ns = _globals)   # Start interactive shell
      #embed.InteractiveShellEmbed.clear_instance()

   else:
      _readline (_locals, _globals)   # Make shell more powerful

      shell = code.InteractiveConsole (namespace () )
      shell.interact ()      # Start interactive shell



##------------------------Start __call__-------------------------------------------##
import sys

class Console (sys.modules[__name__].__class__):
   def __call__(self, _locals = None, _globals = None):  # module callable
      # https://stackoverflow.com/questions/6618795/get-locals-from-calling-namespace-in-python
      if not _locals or not _globals:
         import inspect
         if not _locals:  _locals  = inspect.currentframe ().f_back.f_locals
         if not _globals: _globals = inspect.currentframe ().f_back.f_globals
      shell (_locals, _globals)

sys.modules[__name__].__class__ = Console
