
//---------------------------Class Animation--------------------------------------------//
class Animation {
   constructor (memory, graph) {
      this.memory = memory;
      this.graph = graph;
      this.frameID = undefined;
      this.textGroup = undefined;
      this.lastTimestamp = 0;
      this.updateSpeed = 1000;
      this.haloMaterial = undefined;
      this.cache = {}

      // three.js

      // Load textures (convert URLs into content)
      this.textureLoader = new THREE.TextureLoader ();
      this.textures = {
         //lensflare0_alpha: this.textureLoader.load ('./images/lensflare0_alpha.png'),
         //white: this.textureLoader.load ('./images/white.png')
      };

      // Load fonts (convert URLs into content)
      this.fontLoader = new THREE.FontLoader ();
      this.fonts = {
         droid_sans_regular: new RequestV3rd ("droid_sans_regular.typeface.json").toString ()
      };
      for (var key in this.fonts) if (this.fonts.hasOwnProperty (key) ) this.fontLoader.load (this.fonts[key], font => this.fonts[key] = font);

      // Load shaders (convert URLs into content)
      this.shaders = {
         vertexShader: new RequestV3rd ("vertexShader.c").load (),
         fragmentShader: new RequestV3rd ("fragmentShader.c").load ()
      };
   }//end Constructor


   //------------------------Start initialize-------------------------------------------//
   initialize () {
      try {
         this.initializeCamera ();
         this.initializeHalo ();
      } catch (e) { console.error (e); }

      this.textGroup = new THREE.Group ();
      this.textGroup.name = "Labels";
      this.camera.add (this.textGroup);
   }//end Fct


   //------------------------Start start------------------------------------------------//
   start (frameRequest = requestAnimationFrame) {
      if (this.frameID) this.stop ();

      // this.frameID = frameRequest (this.onAnimate.bind (this) );
      this.frameID = frameRequest (a => this.onAnimate (a) );   // Faster
   }//end Fct


   //------------------------Start stop-------------------------------------------------//
   stop () {
      cancelAnimationFrame (this.frameID);
      cancelAnimationFrame (this.frameID);   // Prevent parallel animation request
      this.frameID = undefined;
   }//end Fct


   //------------------------Getter canvas----------------------------------------------//
   get canvas () {
      return this.graph.canvas;
   }//end Fct


   //------------------------Getter camera----------------------------------------------//
   get camera () {
      return this.graph.camera;
   }//end Fct


   //------------------------Getter scene-----------------------------------------------//
   get scene () {
      return this.graph.scene;
   }//end Fct


   //------------------------Start onAnimate--------------------------------------------//
   onAnimate (timestamp) {
      do { try {
      //if (timestamp - this.lastTimestamp < this.updateSpeed) break;
      //this.lastTimestamp = timestamp;

      //console.log ("Animation.onAnimate()")

      var camera = this.camera;
      this.textGroup.children.forEach (glTextMesh => {
         //var pos = new THREE.Vector3 ().setFromMatrixPosition (glTextMesh.userData.matrixWorld);
         //glTextMesh.position.copy (pos);
         //glTextMesh.lookAt (this.camera.position);
         //glTextMesh.quaternion.copy (this.camera.quaternion);
         glTextMesh.position.setFromMatrixPosition (glTextMesh.userData.__threeObj.matrixWorld);
         glTextMesh.position.project (camera);
         glTextMesh.position.x /= camera.projectionMatrix.elements[0];
         glTextMesh.position.y /= camera.projectionMatrix.elements[5];
         glTextMesh.position.x += 3.4;
         glTextMesh.position.y += 3.4;
         glTextMesh.position.z = -3;
      });
      //camera.updateMatrixWorld ();

   } catch (e) {console.error(e)} } while (false);


   // this.effect.render (this.scene, this.camera);

   // var camera = this.camera;
   // var scene = this.scene;
   // var renderer = this.graph.graph.renderer();

   // scene.updateMatrixWorld();
   // this.stereo.update( camera );

   // renderer.render (scene, this.stereo.cameraL, this.renderTargetL, true);
   // renderer.render( scene, _stereo.cameraR, _renderTargetR, true );
   // renderer.render( _scene, _camera );


   //this.frameID = requestAnimationFrame (this.onAnimate.bind (this) );
   }//end Fct


   //------------------------Start getScreenPosition------------------------------------//
   // https://stackoverflow.com/questions/27409074/converting-3d-position-to-2d-screen-position-r69
   getScreenPosition (obj) {
      var vector = new THREE.Vector3 ();

      var widthHalf = 0.5 * this.canvas.width;
      var heightHalf = 0.5 * this.canvas.height;

      //obj.updateMatrixWorld ();
      vector.setFromMatrixPosition (obj.matrixWorld);
      vector.project (this.camera);

      vector.x = ( vector.x * widthHalf ) + widthHalf;
      vector.y = - ( vector.y * heightHalf ) + heightHalf;

      return vector;
   }//end Fct


   //------------------------Start addBoundingBox---------------------------------------//
   addBoundingBox (bb, rect) {
      var camera = this.camera;
      var widthHalf = 0.5 * this.canvas.width;
      var heightHalf = 0.5 * this.canvas.height;

      var getCameraPos = (x, y, z) => {
         var vector = new THREE.Vector3 ();
         vector.x = (x - widthHalf) / widthHalf;
         vector.y = -(y - heightHalf) / heightHalf;
         vector.x /= camera.projectionMatrix.elements[0];
         vector.y /= camera.projectionMatrix.elements[5];
         vector.z = z;
         return vector;
      };//end Lambda


      var getSize = (x, y) => {
         var vector = new THREE.Vector3 ();
         vector.x = x / widthHalf;
         vector.y = y / heightHalf;
         vector.x /= camera.projectionMatrix.elements[0];
         vector.y /= camera.projectionMatrix.elements[5];
         return vector;
      };//end Lambda


      var getGeometry = () => {
         var target = getSize (rect[2] - rect[0], rect[3] - rect[1]);
         var glPlaneGeo = new THREE.PlaneGeometry (target.x, target.y);
         glPlaneGeo.translate (target.x * 0.5, -target.y * 0.5, 0);
         return glPlaneGeo;
      };//end Lambda


      if (bb && !rect) {
         camera.remove (bb);
      } else if (!bb) {
         var glPlaneGeo = getGeometry ();
         var glPlaneMat = new THREE.MeshBasicMaterial( {color: 0xff0000, side: THREE.DoubleSide, transparent: true, opacity: 0.3, depthWrite: true} );
         bb = new THREE.Mesh (glPlaneGeo, glPlaneMat);
         bb.name = "Bounding Box";

         bb.position.copy (getCameraPos (rect[0], rect[1], -3.2) );

         camera.add (bb);
      } else {
         bb.geometry = getGeometry ();
         bb.position.copy (getCameraPos (rect[0], rect[1], -3.2) );
      }

      return bb;
   }//end Fct


   //------------------------Start addHalo----------------------------------------------//
   addHalo (obj, name = "Halo", color) {
      if (!obj) return;
      if (obj.getObjectByName (name) ) return;
      var halo = new THREE.Mesh (obj.geometry.clone(), this.haloMaterial.clone() );
      halo.name = name;
      halo.scale.multiplyScalar (1.2);
      if (color) halo.material.uniforms.glowColor.value = new THREE.Color (color);
      obj.add (halo);
   }//end Fct


   //------------------------Start removeHalo-------------------------------------------//
   removeHalo (obj, name = "Halo") {
      if (!obj) return;
      obj.remove (obj.getObjectByName (name) );
   }//end Fct


   //------------------------Start resizeObj--------------------------------------------//
   resizeObj (obj, factor) {
      if (!obj || !factor) return;

      if (factor == "reset") {
         obj.scale.x = 1;
         obj.scale.y = 1;
         obj.scale.z = 1;
         return;
      }
      if (factor == -0.4)
      console.log (obj.uuid == "E987A341-BEE0-434F-8365-B214AE742B65")

      obj.scale.addScalar (factor);
   }//end Fct


   //------------------------Start updateFrontText--------------------------------------//
   updateFrontText () {
      var nodes = this.graph.data.nodes;
   }//end Fct


   //------------------------Start addHUDText-------------------------------------------//
   addHUDText (text, opts) {
      var glTextMesh = this._createText (text, opts);

      this.camera.add (glTextMesh);
   }//end Fct


   //------------------------Start addText----------------------------------------------//
   addText (text, opts) {
      var glTextMesh = this._createText (text, opts);

      this.scene.add (glTextMesh);
   }//end Fct


   //------------------------Start addFrontText-----------------------------------------//
   addFrontText (text, opts) {
      //opts.pos = new THREE.Vector3 ().setFromMatrixPosition (opts.object.matrixWorld);
      opts.pos = new THREE.Vector3 ();
      opts.pos.setFromMatrixPosition (opts.node.__threeObj.matrixWorld);
      opts.pos.z = -3;

      var glTextMesh = this._createText (text, opts);
      glTextMesh.userData = opts.node.__threeObj;

      //glTextMesh.onBeforeRender ( function (renderer, scene, camera, geometry, material, group) { material.quaternion.copy (camera.quaternion) } );
      //opts.object.renderOrder = 3;
      //opts.object.material.depthTest = false;
      //opts.node.__threeObj.add (glTextMesh);
      opts.node.__threeLabel = glTextMesh;
      opts.node.__threeLabel.userData = opts.node;
      this.textGroup.add (glTextMesh);
      //this.scene.add (glTextMesh);
      //this.scene.children = [glTextMesh, ...this.scene.children];
      //this.textGroup.add (glTextMesh);
   }//end Fct


   //------------------------Start removeFrontText--------------------------------------//
   removeFrontText (obj) {
      this.textGroup.remove (obj);
   }//end Fct


   //------------------------Start resizeFrontText--------------------------------------//
   resizeFrontText (obj, factor) {
      if (!obj || !factor) return;

      if (factor == "reset") {
         obj.scale.x = 1;
         obj.scale.y = 1;
         return;
      }

      obj.scale.x += factor;
      obj.scale.y += factor;
   }//end Fct


   //------------------------Start _createText------------------------------------------//
   // https://stackoverflow.com/questions/25956272/better-quality-text-in-webgl
   // http://delphic.me.uk/webgltext.html
   _createText (text, opts) {
      var glTextGeo = new THREE.TextGeometry (text, {
         font: this.fonts.droid_sans_regular,
         size: opts.size,
         height: opts.height,
         curveSegments: 0.7,
         bevelEnabled: false
      });

      glTextGeo.computeBoundingBox ();
      glTextGeo.computeVertexNormals ();
      //glTextGeo.center ();

      var glTextMesh = new THREE.Mesh (glTextGeo, new THREE.MeshBasicMaterial ({ color: opts.color, /*depthTest: false*/ }) );

      glTextMesh.position.copy (opts.pos);



      // Background
      if (opts.bgcolor !== false) {
         var glPlaneGeo = new THREE.PlaneGeometry (glTextGeo.boundingBox.max.x - glTextGeo.boundingBox.min.x + 2, glTextGeo.boundingBox.max.y - glTextGeo.boundingBox.min.y + 2);
         var glPlaneMat = new THREE.MeshBasicMaterial( {color: opts.bgcolor || 0xfafafa, side: THREE.DoubleSide, transparent: true, opacity: 0.7, depthWrite: false } );
         var glPlane = new THREE.Mesh (glPlaneGeo, glPlaneMat);
         //glPlaneGeo.center ();
         glPlane.position.x = (glTextGeo.boundingBox.max.x - glTextGeo.boundingBox.min.x) / 2 + 0.5;
         glPlane.position.y = (glTextGeo.boundingBox.max.y - glTextGeo.boundingBox.min.y) / 2 - 0.5;
         glPlane.position.z = -0.1;
         glTextMesh.add (glPlane);
      }



      return glTextMesh;
   }//end Fct


   //------------------------Start setCamera--------------------------------------------//
   setCamera (mode) {
      var interaction = true;
      switch (mode) {
      case "orthographic": {
         this.graph.graph.render (this.cache.render);
         this.graph.graph.camera (this.cache.orthographic);
      } break;
      case "perspective": {
         this.graph.graph.render (this.cache.render);
         this.graph.graph.camera (this.cache.perspective);
      } break;
      case "anaglyph": {
         // Stereo Effect
         this.graph.graph.render ( (s,c) => this.cache.anaglyph.render (s,c) );
         this.graph.graph.camera (this.cache.perspective);
      } break;
      case "parallax": {
         // Stereo Effect
         this.graph.graph.render ( (s,c) => this.cache.parallax.render (s,c) );
         this.graph.graph.camera (this.cache.perspective);
      } break;
      case "sidebyside": {
         // Stereo Effect
         this.graph.graph.render ( (s,c) => this.cache.sidebyside.render (s,c) );
         this.graph.graph.camera (this.cache.perspective);
         interaction = false;
      } break;
      case "vr": {

      } break;
      default: return;
      };//end switch

      this.graph.controls.camMode = mode;
      this.scene.children.forEach (obj => { if (obj.isCamera) this.scene.remove (obj) });
      // this.camera.add (this.textGroup);
      this.scene.add (this.camera);
      this.graph.graph.height (this.graph.graph.height () ).width (this.graph.graph.width () );
      this.graph.graph.enablePointerInteraction (interaction);
   }//end Fct


   //------------------------Start initializeCamera-------------------------------------//
   initializeCamera () {
      var scene = this.scene;

      // Delete directional light
      scene.children.some (function (e, p, a) { if (e.type == "DirectionalLight") { a.splice(p, 1); return true; } return false; });

      // Store default Objects
      this.cache.render = this.graph.graph.render ();
      this.cache.perspective = this.graph.graph.camera ();

      // Orthographic camera
      var orthographic = this.cache.orthographic = new THREE.OrthographicCamera (window.innerWidth / 2, window.innerWidth / -2, window.innerHeight / -2, window.innerHeight / 2, 1, 1000);
      orthographic.position.z = 1000;
      orthographic.far = 50000;
      // Use callbacks for the camera size
      orthographic.userData = this.graph.graph;
      Object.defineProperty (orthographic, "left", { get: function () { return - this.userData.width() / 2 } });
      Object.defineProperty (orthographic, "right", { get: function () { return this.userData.width() / 2 } });
      Object.defineProperty (orthographic, "top", { get: function () { return this.userData.height() / 2 } });
      Object.defineProperty (orthographic, "bottom", { get: function () { return - this.userData.height() / 2 } });
      // Zoom a little bit
      orthographic.zoom = 3;

      // Stereo Effect (Anaglyph)
      this.cache.anaglyph = new THREE.AnaglyphEffect (this.graph.graph.renderer (), this.graph.graph.width(), this.graph.graph.height() );
      // this.cache.anaglyph._stereo.eyeSep = 0.064;

      // Stereo Effect (ParallaxBarrier)
      this.cache.parallax = new THREE.ParallaxBarrierEffect (this.graph.graph.renderer () );
      // this.cache.parallax._stereo.eyeSep = 0.064;   //TODO patch library

      // Stereo Effect (Side-by-side)
      this.cache.sidebyside = new THREE.StereoEffect (this.graph.graph.renderer () );
      // this.cache.sidebyside._stereo.eyeSep = 0.064;
      this.cache.sidebyside.setEyeSeparation (0.05);

      // Flash Light
      var light = new THREE.DirectionalLight (0xffffff, 0.4);
      light.position.set(0, 100, 0);
      this.cache.perspective.add (light.clone () );
      this.cache.orthographic.add (light.clone () );

      // Controls
      var controls = this.graph.controls;
      controls.minDistance = 0;
      controls.maxDistance = 50000;
      controls.staticMoving = true;
      controls.panSpeed = 0.1;
      controls.camMode = "orthographic";

      // Replace camera
      scene.add (this.graph.graph.camera (this.cache.orthographic) );
   }//end Fct


   //------------------------Start initializeHalo---------------------------------------//
   initializeHalo () {
      this.haloMaterial = new THREE.ShaderMaterial ({
         uniforms: {
            "c":   { type: "f", value: 0.9 },
            "p":   { type: "f", value: 2 },
            glowColor: { type: "c", value: new THREE.Color (0xffffff) }
         },
         vertexShader:   this.shaders.vertexShader,
         fragmentShader: this.shaders.fragmentShader,
         side: THREE.FrontSide,
         // side: THREE.BackSide,
         blending: THREE.AdditiveBlending,
         transparent: true
      });
   }//end Fct
};//end Class
