

//---------------------------Prototype forceFR------------------------------------------//
// https://bl.ocks.org/starcalibre/71409a4aab6c85bc13d9fc5a4fd4d7af
// https://github.com/d3/d3-force/blob/master/src/link.js
function forceFR (fmmm, scale, gravity) {
   var nodes = [];
   var links = [];
   var id;

   function fAttr(x,k) { return (x * x) / k; }
   function fRepl(x,k) { return (k * k) / x; }
   function clamp(n, min, max) { return Math.min(Math.max(n, min), max); }
   function subVectors(a, b) { return new THREE.Vector3 (a.x - b.x, a.y - b.y, a.z - b.z); }
   function vecLength(v) { return Math.sqrt( v.x * v.x + v.y * v.y + v.z * v.z ); }


   //------------------------FR force---------------------------------------------------//
   function force (alpha) {
      if (!links || !links.length || typeof (links[0].source) == "string") return;

      var width = 2 * fmmm;
      var height = 2 * fmmm;
      var depth = 2 * fmmm;

      var area = scale * nodes.length * nodes.length;
      var k = Math.pow(area / nodes.length + 1, 1 / 3);


      var t = Math.sqrt(nodes.length);
      // var dt = t / (iter + 1);
      var eps = 0.05;

      // calculate repulsive forces O(N^2)
      nodes.forEach( (v, p1) => {
         v.disp = new THREE.Vector3();

         nodes.forEach( (u, p2) => {
            if (p1 == p2) return true;

            var delta = subVectors(v, u);
            var deltaMag = Math.max(eps, delta.length());

            v.disp.x += (delta.x / deltaMag) * fRepl(deltaMag, k);
            v.disp.y += (delta.y / deltaMag) * fRepl(deltaMag, k);
            v.disp.z += (delta.z / deltaMag) * fRepl(deltaMag, k);
         });//end for 3
      });//end for 2

      // calculate attractive forces O(L)
      links.forEach( (e, i) => {
         var delta = subVectors (e.source, e.target);
         var deltaMag = Math.max(eps, delta.length());

         e.source.disp.x -= (delta.x / deltaMag) * fAttr(deltaMag, k);
         e.source.disp.y -= (delta.y / deltaMag) * fAttr(deltaMag, k);
         e.source.disp.z -= (delta.z / deltaMag) * fAttr(deltaMag, k);
         e.target.disp.x += (delta.x / deltaMag) * fAttr(deltaMag, k);
         e.target.disp.y += (delta.y / deltaMag) * fAttr(deltaMag, k);
         e.target.disp.z += (delta.z / deltaMag) * fAttr(deltaMag, k);
      });//end for 2

      // apply gravitational forces O(N)
      nodes.forEach( (v, i) => {
         // three.js co-ordinate space is the center, so we
         // just need the regular magnitude for each vertices position
         var deltaMag = Math.max(eps, vecLength (v) );
         var gravityForce = 0.1 * k * gravity * deltaMag;

         v.disp.x -= (v.x / deltaMag) * gravityForce;
         v.disp.y -= (v.y / deltaMag) * gravityForce;
         v.disp.z -= (v.z / deltaMag) * gravityForce;
      });//end for 2

      // ??? O(N)
      nodes.forEach(v => {
         // limit displacement to temperature t
         var dispMag = vecLength(v.disp);
         v.disp.x += (v.disp.x / dispMag) * Math.min(Math.abs(v.disp.x), t);
         v.disp.y += (v.disp.y / dispMag) * Math.min(Math.abs(v.disp.y), t);
         v.disp.z += (v.disp.z / dispMag) * Math.min(Math.abs(v.disp.z), t);

         // keep vertices within the world space
         v.disp.x = clamp(v.disp.x, -fmmm + width / 10, fmmm - width / 10);
         v.disp.y = clamp(v.disp.y, -fmmm + height / 10, fmmm - height / 10);
         v.disp.z = clamp(v.disp.z, -fmmm + depth / 10, fmmm - depth / 10);

         // The velocity is the distance between old and new position
         v.vx = (v.disp.x - v.x) * 0.1 * alpha;
         v.vy = (v.disp.y - v.y) * 0.1 * alpha;
         v.vz = (v.disp.z - v.z) * 0.1 * alpha;

         // Force the new position
         // v.x = v.disp.x;
         // v.y = v.disp.y;
         // v.z = v.disp.z;
         // v.vx = 0.2;
         // v.vy = 0.2;
         // v.vz = 0.2;
      });//end for 2

      // t -= dt;
   }//end Fct


   force.initialize = function(_) {
      nodes = _;
   };//end Fct

   force.scale = function(_) {
      return arguments.length ? (scale = +_, force) : scale;
   };//end Fct

   force.gravity = function(_) {
      return arguments.length ? (gravity = +_, force) : gravity;
   };//end Fct



   //------------------------Link force-------------------------------------------------//
   function index(d) { return d.id; }
   function find(nodeById, nodeId) {
      // console.log (nodeById)
      // console.log (nodeId)
      var node = nodeById.get(nodeId);
      if (!node) throw new Error("missing: " + nodeId);
      return node;
   }
   function defaultStrength(link) {
      return 1 / Math.min(count[link.source.index], count[link.target.index]);
   }
   function constant(x) {
      return function() {
         return x;
      };
   }

   var id = index,
      strength = defaultStrength,
      strengths,
      distance = constant(30),
      distances,
      count,
      bias,
      iterations = 1;

   function link (alpha) {}


   function initialize() {
      if (!nodes || !nodes.length) return;
      if (!links || !links.length) return;

      var i,
          n = nodes.length,
          m = links.length,
          nodeById = d3.map(nodes, id),
          link;
window.nodeById = nodeById
// console.log (nodes)
// console.log (id)
// console.log (links)
      for (i = 0, count = new Array(n); i < m; ++i) {
         link = links[i], link.index = i;
         if (typeof link.source !== "object") link.source = find(nodeById, link.source);
         if (typeof link.target !== "object") link.target = find(nodeById, link.target);
         count[link.source.index] = (count[link.source.index] || 0) + 1;
         count[link.target.index] = (count[link.target.index] || 0) + 1;
      }

      for (i = 0, bias = new Array(m); i < m; ++i) {
         link = links[i], bias[i] = count[link.source.index] / (count[link.source.index] + count[link.target.index]);
      }

      strengths = new Array(m), initializeStrength();
      distances = new Array(m), initializeDistance();
   }

   function initializeStrength() {
      if (!nodes) return;

      for (var i = 0, n = links.length; i < n; ++i) {
         strengths[i] = +strength(links[i], i, links);
      }
   }

   function initializeDistance() {
      if (!nodes) return;

      for (var i = 0, n = links.length; i < n; ++i) {
         distances[i] = +distance(links[i], i, links);
      }
   }

   link.initialize = function(_) {
      nodes = _;
      initialize();
   };

   link.links = function(_) {
      // console.log (_)
      return arguments.length ? (links = _, initialize(), link) : links;
   };

   link.id = function(_) {
      return arguments.length ? (id = _, link) : id;
   };

   link.iterations = function(_) {
      return arguments.length ? (iterations = +_, link) : iterations;
   };

   link.strength = function(_) {
      return arguments.length ? (strength = typeof _ === "function" ? _ : constant(+_), initializeStrength(), link) : strength;
   };

   link.distance = function(_) {
      return arguments.length ? (distance = typeof _ === "function" ? _ : constant(+_), initializeDistance(), link) : distance;
   };

   return { fr: force, link: link };
}//end Prototype
