
//---------------------------Class Graph------------------------------------------------//
class Graph {
   constructor (memory, node) {
      this.memory = memory;
      this.node = node;
      this.graph = undefined;
      this.animation = new Animation (this.memory, this);
      this.selected = undefined;
      this.allData = undefined;
      this.lastClick = undefined;
      this.request = undefined;
      this.config = { resolution: 12, linkWidth: 0.8, nodeSize: 1, nodeOpacity: 1, linkOpacity: 0.5, warmup: 10, cooldown: 99999, cooldownTime: 60000, forceEngine: 'd3', d3VelocityDecay: 0.6, d3AlphaDecay: 0.016, d3Layout: undefined };
      this.sconfig = {};
      this.forces = ['link', 'center', 'charge', 'collide'];
      this.fr = [2000, 25, 1];
      //this.keys = { alt: false, ctrl: false, shift: false };
      this.keys = {};
      this.bbSelect = undefined;
      this.bbTimeStamp = 0;
      this.colors = { AgeingFactor_gene: "#FF00FF", AgeingFactor_compound: "#C89BC8", AgeingFactor_other_factor: "#B05DE5", AlleleType: "#F4A460", Citation: "#D2B48C", GeneOntologyTerm_Component: "#48D1CC", GeneOntologyTerm_Function: "#87CDFA", GeneOntologyTerm_Process: "#00BFFF", KEGGPathwayExtraGenes: "#C28F8F", Observation: "#00FF00", Observation_dec: "#FF0000", Observation_inc: "#9ACD32", Observation_lifespan: "#808080", Observation_homology_analysis: "#0000FF", Pathway: "#AB5C00", Species: "#F0F041", default: "yellow", bg: "#000011" };
      this.values = { AgeingFactor: 25, AlleleType: 18, Citation: 17, GeneOntologyTerm: 10, KEGGPathwayExtraGenes: 14, Observation: 15, Pathway: 30, Species: 35, default: 5 };
      this.active = { AgeingFactor: true, AlleleType: true, Citation: true, GeneOntologyTerm: true, KEGGPathwayExtraGenes: true, Observation: true, Pathway: true, Species: true, GOI: false };
      this.mapping = { AF: "AgeingFactor", AT: "AlleleType", CT: "Citation", GO: "GeneOntologyTerm", KEGGX: "KEGGPathwayExtraGenes", OB: "Observation", PW: "Pathway", SP: "Species" };


      // Mouse keystrokes listener
      this.node.addEventListener ("click", event => this.keys = event, true);
      //this.node.addEventListener ("click", this.onClick.bind (this), true);
      this.node.addEventListener ("mousedown", this.onMouseDown.bind (this), true);

      this.registerResize();
   }//end Constructor


   //------------------------Getter canvas----------------------------------------------//
   get canvas () {
      return this.node.getElementsByTagName("canvas")[0];
   }//end Fct


   //------------------------Getter camera----------------------------------------------//
   get camera () {
      return this.graph.camera ();
   }//end Fct


   //------------------------Getter controls--------------------------------------------//
   get controls () {
      return this.graph.controls ();
   }//end Fct


   //------------------------Getter scene-----------------------------------------------//
   get scene () {
      return this.graph.scene ();
   }//end Fct


   //------------------------Getter renderer--------------------------------------------//
   get renderer () {
      return this.graph.renderer ();
   }//end Fct


   //------------------------Getter data------------------------------------------------//
   get data () {
      return this.graph.graphData ();
   }//end Fct


   //------------------------Setter data------------------------------------------------//
   set data (data_) {
      this.graph.graphData (data_);
      this.update ();
   }//end Fct


   //------------------------Start update-----------------------------------------------//
   update () {
      this.graph.nodeRelSize (this.graph.nodeRelSize () );

      setTimeout ( () => {
         this.data.nodes.forEach (node => {
            if (node._selected) this.animation.addHalo (node.__threeObj, "Selected");
            if (node._goi && this.active.GOI) this.animation.addHalo (node.__threeObj, "GOI", this.colors.AgeingFactor_gene);
         });
      }, 50);

      //this.graph.updateFrontText ();
      // setTimeout ( () => {
      //    //this.data.nodes.forEach (node => { if (node.__threeLabel) node.__threeObj.add (node.__threeLabel) });
      //    //this.data.nodes.forEach (node => { if (node.__threeLabel) node.__threeLabel.userData = node });
      // }, 1000);
   }//end Fct


   //------------------------Start updateLayout-----------------------------------------//
   updateLayout (ticks = 200) {
      this.data = this.data;
      this.graph
         .warmupTicks (0)
         .cooldownTime (this.config.cooldownTime)
         .cooldownTicks (ticks);
   }//end Fct


   //------------------------Start stopLayout-------------------------------------------//
   stopLayout () {
      this.graph
         .cooldownTicks (0)
         .cooldownTime (0);
   }//end Fct


   //------------------------Start updateColors-----------------------------------------//
   updateColors () {
      this.graph.backgroundColor (this.memory.graph.colors.bg);
      this.data.nodes.forEach (node => {
         if (node.__threeLabel)
            node.__threeLabel.material.color.set (this.nodeColor (node) );
      });
      this.update ();
      this.memory.hover.showLegend ();
   }//end Fct


   //------------------------Start update-----------------------------------------------//
   freeze () {
      this.animation.stop ();
      this.graph.stopAnimation ();
   }//end Fct


   //------------------------Start createGraph------------------------------------------//
   createGraph (request, callback, backend = '3D') {
      this.memory.statusbar.hide ();
      this.memory.spinner.show ("Loading data ...");
      if (this.graph) this.animation.setCamera ("orthographic");   // Reset camera

      // var onerror = e => this.memory.statusbar.error = "Error loading graph: " + e.toString ();
      var onerror = e => this.memory.statusbar.error = e;

      var createGraph_ = (data) => {
         if (data.error !== undefined) { throw data.error; }
         this.memory.spinner.show ("Creating graph ...");
         switch (backend) {
         case "2D": this.createGraph2D (data); break;
         case "3D": this.createGraph3D (data); break;
         case "VR": this.createGraphVR (data); break;
         case "THREE": this.createGraphThree (data); break;
         };//end switch
         this.modifyGraph (callback);
      };//end Lambda

      try {
         if (request instanceof Request) {
            request.catch (onerror).async (createGraph_);
         } else if (typeof (request) == "string") {
            fetch (request).catch (onerror)
               .then (res => res.json () ).catch (onerror)
               .then (createGraph_).catch (onerror);
         } else if (typeof (request) == "object" && request.nodes !== undefined && request.links !== undefined) {
            try { createGraph_ (request); } catch (e) { onerror (e); console.error (e); }
         } else {
            throw "cannot process request";
         }
      } catch (e) { onerror (e); console.error (e); }

      this.request = request;
   }//end Fct


   //------------------------Start createGraphVR----------------------------------------//
   createGraphVR (data) {
      if (!window.ForceGraphVR) loadScript (new Request().viewer + "/3rdParty/3d-force-graph-vr.min.js", null, null, false);

      var config = this.sconfig = this.mergeConfig (data.config, data);

      this.graph = ForceGraphVR()(this.node)
         .graphData(data)
         .height(parseInt(window.getComputedStyle(this.node.parentNode)['height'].split('px')[0]))
         .width(parseInt(window.getComputedStyle(this.node.parentNode)['width'].split('px')[0]))
         .linkWidth(config.linkWidth)
         .linkOpacity(config.linkOpacity)
         .nodeRelSize(config.nodeSize)
         .nodeOpacity(config.nodeOpacity)
         .nodeResolution(config.resolution)
         .enableNodeDrag(false)
         .backgroundColor(this.colors.bg)
         .nodeColor(this.nodeColor.bind (this) )
         .nodeVal(this.nodeValue.bind (this) )
         .linkColor(this.linkColor.bind (this) )
         .onNodeClick(this.onNodeClick.bind (this) )
         .onLinkClick(this.onLinkClick.bind (this) )
         .onNodeHover(this.onNodeHover.bind (this) )
         .onLinkHover(this.onLinkHover.bind (this) )
         .warmupTicks(config.warmup)
         .cooldownTicks(config.cooldown)
         .forceEngine(config.forceEngine)
         .nodeAutoColorBy('type')
         .linkAutoColorBy('type')
         .nodeThreeObject(undefined);

         if (config.cooldown == this.config.cooldown)
            this.graph.cooldownTime(config.cooldownTime)
   }//end Fct


   //------------------------Start createGraph3D----------------------------------------//
   createGraph3D (data) {
      if (!window.ForceGraph3D) loadScript (new Request().viewer + "/3rdParty/3d-force-graph.min.js", null, null, false);

      var config = this.sconfig = this.mergeConfig (data.config, data);

      this.graph = ForceGraph3D()(this.node)
         .graphData(data)
         .height(parseInt(window.getComputedStyle(this.node.parentNode)['height'].split('px')[0]))
         .width(parseInt(window.getComputedStyle(this.node.parentNode)['width'].split('px')[0]))
         .linkWidth(config.linkWidth)
         .linkOpacity(config.linkOpacity)
         .nodeRelSize(config.nodeSize)
         .nodeOpacity(config.nodeOpacity)
         .nodeResolution(config.resolution)
         .enableNodeDrag(false)
         .backgroundColor(this.colors.bg)
         .nodeLabel(this.nodeLabel.bind (this) )
         .nodeColor(this.nodeColor.bind (this) )
         .nodeVal(this.nodeValue.bind (this) )
         .linkColor(this.linkColor.bind (this) )
         .onNodeClick(this.onNodeClick.bind (this) )
         .onLinkClick(this.onLinkClick.bind (this) )
         .onNodeHover(this.onNodeHover.bind (this) )
         .onLinkHover(this.onLinkHover.bind (this) )
         .warmupTicks(config.warmup)
         .cooldownTicks(config.cooldown)
         .forceEngine(config.forceEngine)
         .nodeAutoColorBy('type')
         .linkAutoColorBy('type')
         .enablePointerInteraction(true)
         .enableNavigationControls(true)
         //.nodeThreeObject (undefined);

      if (config.cooldown == this.config.cooldown)
         this.graph.cooldownTime(config.cooldownTime)

      this.d3ForceLayout (config);
   }//end Fct


   //------------------------Start createGraph2D----------------------------------------//
   createGraph2D (data) {
      if (!window.ForceGraph) loadScript (new Request().viewer + "/3rdParty/force-graph.min.js", null, null, false);

      var config = this.sconfig = this.mergeConfig (data.config, data);

      this.graph = ForceGraph()(this.node)
         .graphData(data)
         .height(parseInt(window.getComputedStyle(this.node.parentNode)['height'].split('px')[0]))
         .width(parseInt(window.getComputedStyle(this.node.parentNode)['width'].split('px')[0]))
         .linkWidth(config.linkWidth)
         //.linkOpacity(config.linkOpacity)
         .nodeRelSize(config.nodeSize)
         //.nodeOpacity(config.nodeOpacity)
         //.nodeResolution(config.resolution)
         .enableNodeDrag(false)
         .backgroundColor(this.colors.bg)
         .nodeLabel(this.nodeLabel.bind (this) )
         .nodeColor(this.nodeColor.bind (this) )
         .nodeVal(this.nodeValue.bind (this) )
         .linkColor(this.linkColor.bind (this) )
         .onNodeClick(this.onNodeClick.bind (this) )
         .onLinkClick(this.onLinkClick.bind (this) )
         .onNodeHover(this.onNodeHover.bind (this) )
         .onLinkHover(this.onLinkHover.bind (this) )
         .warmupTicks(config.warmup)
         .cooldownTicks(config.cooldown)
         //.forceEngine(config.forceEngine)
         //.d3Force(config.d3Force || 'link')
         //.nodeAutoColorBy('type')
         //.linkAutoColorBy('type')
         //.nodeThreeObject(undefined);


         if (config.cooldown == this.config.cooldown)
            this.graph.cooldownTime(config.cooldownTime)
      }//end Fct


   //------------------------Start createGraphThree-------------------------------------//
   createGraphThree (data) {
      if (!window.ThreeForceGraph) loadScript (new Request().viewer + "/3rdParty/three-forcegraph.min.js", null, null, false);
      // if (!window.TrackballControls) loadScript (new Request().viewer + "/3rdParty/three-trackballcontrols.min.js", null, null, false);

      var config = this.sconfig = this.mergeConfig (data.config, data);

      this.graph = new ThreeForceGraph()
         .graphData(data)
         //.height(parseInt(window.getComputedStyle(this.node.parentNode)['height'].split('px')[0]))
         //.width(parseInt(window.getComputedStyle(this.node.parentNode)['width'].split('px')[0]))
         .linkWidth(config.linkWidth)
         .linkOpacity(config.linkOpacity)
         .nodeRelSize(config.nodeSize)
         .nodeOpacity(config.nodeOpacity)
         .nodeResolution(config.resolution)
         //.enableNodeDrag(false)
         //.backgroundColor(this.colors.bg)
         //.nodeLabel(this.nodeLabel.bind (this) )
         .nodeColor(this.nodeColor.bind (this) )
         .nodeVal(this.nodeValue.bind (this) )
         .linkColor(this.linkColor.bind (this) )
         //.onNodeClick(this.onNodeClick.bind (this) )
         //.onLinkClick(this.onLinkClick.bind (this) )
         //.onNodeHover(this.onNodeHover.bind (this) )
         //.onLinkHover(this.onLinkHover.bind (this) )
         .warmupTicks(config.warmup)
         .cooldownTicks(config.cooldown)
         .forceEngine(config.forceEngine)
         .nodeAutoColorBy('type')
         .linkAutoColorBy('type')
         //.enablePointerInteraction(true)
         //.enableNavigationControls(true)
         //.nodeThreeObject (undefined);
         .d3VelocityDecay (0.8)

         if (config.cooldown == this.config.cooldown)
         this.graph.cooldownTime(config.cooldownTime)


      // Setup renderer
      const renderer = new THREE.WebGLRenderer();
      // renderer.setSize (
      //    parseInt(window.getComputedStyle(this.node.parentNode)['width'].split('px')[0]),
      //    parseInt(window.getComputedStyle(this.node.parentNode)['height'].split('px')[0])
      // );
      renderer.setSize (
         window.innerWidth,
         window.innerHeight
      );
      this.node.appendChild (renderer.domElement);

      // Setup scene
      const scene = new THREE.Scene ();
      scene.add (this.graph);
      scene.add (new THREE.AmbientLight (0xbbbbbb) );

      // Setup camera
      //const camera = new THREE.PerspectiveCamera ();
      const camera = new THREE.OrthographicCamera (window.innerWidth / - 2, window.innerWidth / 2, window.innerHeight / 2, window.innerHeight / - 2, 1, 1000);
      camera.aspect = window.innerWidth / window.innerHeight;
      camera.updateProjectionMatrix ();
      camera.lookAt (this.graph.position);
      camera.position.z = Math.cbrt(300) * 100;

      // var camera = new THREE.StereoCamera ();
      // camera.aspect = 1;

      // camera.eyeSep = 0.064;
      // // camera.updateProjectionMatrix ();
      // //camera.lookAt (this.graph.position);
      // //camera.position.z = Math.cbrt(300) * 100;

      // camera.cameraL = new THREE.OrthographicCamera();
      // camera.cameraL.layers.enable( 1 );
      // camera.cameraL.matrixAutoUpdate = false;

      // camera.cameraR = new THREE.OrthographicCamera();
      // camera.cameraR.layers.enable( 2 );
      // camera.cameraR.matrixAutoUpdate = false;

      var light = new THREE.DirectionalLight(0xffffff, 0.4);
      light.position.set(0, 100, 0);

      scene.add(camera);
      camera.add(light);

      // Add camera controls
      const tbControls = new TrackballControls (camera, renderer.domElement);
      // const tbControls = new THREE.OrbitControls (camera);


      // Kick-off renderer
      var animate = () => { // IIFE
         this.graph.tickFrame();
         // Frame cycle
         tbControls.update();
         renderer.render (scene, camera);
         requestAnimationFrame (animate);
      };

      animate.bind(this)();
   }//end Fct


   //------------------------Start d3ForceLayout----------------------------------------//
   d3ForceLayout (config) {
      if (!config) config = this.config;
      if (typeof (config) == "string") config = this.mergeConfig ({d3Layout: config});
      this.graph
         .d3VelocityDecay (config.d3VelocityDecay)
         .d3AlphaDecay (config.d3AlphaDecay)

      switch (config.d3Layout) {
      case "FR": {
         // Deactivate existing forces
         this.forces.forEach (force => this.graph.d3Force (force, null) );
         var { fr, link } = forceFR (this.fr[0], this.fr[1], this.fr[2]);
         this.graph.d3Force ('link', link);
         this.graph.d3Force ('link').links (this.data.links);
         this.graph.d3Force ('fr', fr);
         this.forces = ['fr', 'link'];
      } break;
      case "BB": {
         // Deactivate existing forces
         this.forces.forEach (force => this.graph.d3Force (force, null) );

         // Add collision and bounding box forces
         this.graph.d3Force ('collide', d3.forceCollide (this.graph.nodeRelSize() * 3) );   // Overlapping of nodes
         this.graph.d3Force ('box', () => {
            const CUBE_HALF_SIDE = Graph.nodeRelSize() * N * 0.5;
            nodes.forEach(node => {
               const x = node.x || 0, y = node.y || 0, z = node.z || 0;
               // bounce on box walls
               if (Math.abs(x) > CUBE_HALF_SIDE) { node.vx *= -1; }
               if (Math.abs(y) > CUBE_HALF_SIDE) { node.vy *= -1; }
               if (Math.abs(z) > CUBE_HALF_SIDE) { node.vz *= -1; }
            });
         });
         this.forces = ['collide', 'box'];
      } break;
      case "FMMM": {
         this.memory.spinner.show ("Generating FMMM layout ...");
         var links = [];
         this.data.links.forEach (link => links.push ({source: link.source.id, target: link.target.id}));
         RequestSpecial ("type=layout&layout=fmmm", {method: 'POST', data: JSON.stringify (links)}).catch (e => this.memory.statusbar.error = e).async (data => {
            if (data.error) { this.memory.statusbar.error = data.error;  return; }
            this.data.nodes.forEach (node => {
               var n = data.nodes[node.id];
               if (!n) return;
               node.x = n.x; node.y = n.y; node.z = n.z;
               node.vx = 0; node.vy = 0; node.vz = 0;
            });//end for
            this.graph
               .warmupTicks(config.warmup)
               .cooldownTicks(config.cooldown)
               .cooldownTime (0);
            this.memory.spinner.hide ();
            this.update ();
         });
      } break;
      case "default": {
         // Deactivate existing forces
         this.forces.forEach (force => this.graph.d3Force (force, null) );
         this.graph.d3Force ('link', d3.forceLink () );
         this.graph.d3Force ('charge', d3.forceManyBody () );
         this.graph.d3Force ('center', d3.forceCenter () );
         this.forces = ['link', 'center', 'charge', 'collide'];
         this.data.nodes.forEach (node => {delete node.x; delete node.vx; delete node.y; delete node.vy; delete node.z; delete node.vz; });
      }
      default:
         // Spread nodes a little wider
         this.graph.d3Force ('link').distance (20);
         this.graph.d3Force ('link').iterations (3);
         //this.graph.d3Force ('charge').distanceMax (50);
         this.graph.d3Force ('charge').strength (-30);
         //this.graph.d3Force ('center').x (50);
         this.graph.d3Force ('collide', d3.forceCollide (this.graph.nodeRelSize() * 3) );   // Overlapping of nodes
      };//end switch
   }//end Fct


   //------------------------Start mergeConfig------------------------------------------//
   mergeConfig (config, data) {
      if (data) {            //FIXME better integration
         var networkSizeFactor = Math.max (1, Math.ceil (Math.log10 (data.nodes.length + 1) ) );
         this.config.cooldown = Math.pow (4, networkSizeFactor) * 2 - this.config.warmup;
         // this.config.warmup = Math.pow (6, networkSizeFactor);
         // this.config.cooldown = this.config.warmup;
      }

      if (config === undefined) return this.config;

      if (config.colors) {
         for (var key in config.colors)
            if (this.colors.hasOwnProperty (key) )
               this.colors[key] = config.colors[key];

         this.memory.hover.showLegend ();
      }

      if (config.values)
         for (var key in config.values)
            if (this.values.hasOwnProperty (key) )
               this.values[key] = config.values[key];

      if (config.active) {
         for (var key in config.active)
            if (this.active.hasOwnProperty (key) )
               this.active[key] = config.active[key];

         this.memory.hover.showLegend ();
      }


      return Object.assign (this.config, config);
   }//end Fct


   //------------------------Start modifyGraph------------------------------------------//
   modifyGraph (callback) {
      this.resize (
         parseInt(window.getComputedStyle(this.node.parentNode)['height'].split('px')[0]),
         parseInt(window.getComputedStyle(this.node.parentNode)['width'].split('px')[0])
      );

      this.animation.initialize ();

      //setTimeout (this.generateNodeLinks.bind (this), 1000);
      new Promise ( (resolve, reject) => {
         setTimeout ( () => { this.generateNodeLinks (); resolve (); }, 350);
      }).then ( () => {
         this.memory.spinner.hide ();
         this.animation.start (this.graph.tick);
         if (callback) callback ();
      });
   }//end Fct


   //------------------------Start generateNodeLinks------------------------------------//
   generateNodeLinks () {
      var { nodes, links } = this.data;
      this.allData = { nodes: [...nodes], links: [...links] };

      nodes.forEach (node => {
         node.neighbors = [];
         node.links = [];

         if (node._req || node.type == "Species") this.addNodeLabel (node);   // Labels
         // if (node._req && this.active.GOI) this.animation.addHalo (node.__threeObj, "GOI", this.colors.AgeingFactor_gene);   // GOI Halo
      });//end for

      links.forEach (link => {
         link.source.neighbors = [...link.source.neighbors, link.target];
         link.target.neighbors = [...link.target.neighbors, link.source];
         link.source.links = [...link.source.links, link];
         link.target.links = [...link.target.links, link];
      });//end for

      // Get max lifespan change
      try {
         nodes.forEach (node => {
            if (node.type == "AgeingFactor") {
               node.lifespanChange;
               var max_ = 0;
               node.links.forEach (link => {
                  if (link.target.type == "Observation" && link.target.subtype == "lifespan" && link.target.lifespanEffectSignificant != "none" && link.target.lifespanChange != 0) {
                     var abs_ = Math.abs (parseFloat (link.target.lifespanChange) );
                     if (abs_ > max_) {
                        max_ = abs_;
                        node.lifespanChange = parseFloat (link.target.lifespanChange);
                     }
                  }
               });//end for 2
            } else if (node.lifespanChange === null) {
               node.lifespanChange = 0;
            }
         });//end for 1
      } catch (e) { console.error (e); }

      this.update ();
   }//end Fct


   //------------------------Start nodeLabel--------------------------------------------//
   nodeLabel (node) {
      if (node.name) return node.name;
      if (node.term) return node.term;
      if (node.symbol) return node.symbol;
      if (node.type == "Species") return node.id;
   }//end Fct


   //------------------------Start nodeColor--------------------------------------------//
   nodeColor (node) {
      if (node._scale && node.__threeObj) node.__threeObj.scale.copy (node._scale);
      if (!this.active[node.type]) return "transparent";
      if (node._color) return node._color;
      if (node.color) return node.color;

      var type = node.type;

      switch (node.type) {
      case "AgeingFactor": {
         type = node.type + "_" + node.subtype;
         //if (node._req) this.animation.addFrontText (node.name, {color: this.colors[type], size: 4, height: 1, object: node.__threeObj});
      } break;
      case "Observation": {
         type = node.type;
         if (node.subtype == 'lifespan') {
            if (node.lifespanEffectSignificant == "increased")
               type += "_inc";
            else if (node.lifespanEffectSignificant == "decreased")
               type += "_dec";
            else
               type += "_lifespan";
         } else
            type += "_" + node.subtype;
      } break;
      case "GeneOntologyTerm": {
         var type = node.type + "_" + node.category;
      } break;
      };//end switch

      if (this.colors.hasOwnProperty(type)) return this.colors[type];
      else return this.colors.default;
   }//end Fct


   //------------------------Start linkColor--------------------------------------------//
   linkColor (link) {
      var spl = link._type.split ('-');
      if (!this.active[this.mapping[spl[0]]]) return "transparent";
      if (!this.active[this.mapping[spl[1]]]) return "transparent";
      if (link._color) return link._color;
      if (link.color) return link.color;

      switch (link._type) {
      case "AF-OB": {
         if (typeof (link.target) == "object")
            return this.nodeColor (link.target);
         else
            return this.colors.Observation_lifespan;
      }
      case "AF-AT": return this.colors.AlleleType;
      case "AF-SP": return this.colors.Species;
      case "AF-PW": return this.colors.Pathway;
      case "AF-GO": return this.colors.GeneOntologyTerm_Component;
      case "OB-SP": return this.colors.Species;
      case "OB-AT": return this.colors.AlleleType;
      case "OB-CT": return this.colors.Citation;
      case "PW-SP": return this.colors.Species;
      case "PW-KEGGX": return this.colors.KEGGPathwayExtraGenes;
      case "KEGGX-GO": return this.colors.GeneOntologyTerm_Component;
      default: return this.colors.default;
      };//end switch
   }//end Fct


   //------------------------Start nodeValue--------------------------------------------//
   nodeValue (node) {
      if (node.val) return node.val;
      if (this.values.hasOwnProperty(node.type)) return this.values[node.type];
      else return this.values.default;
   }//end Fct


   //------------------------Start onNodeClick------------------------------------------//
   onNodeClick (node) {
      if (this.keys.timeStamp - this.bbTimeStamp <= 10) {   // If BBox Select
      } else if (this.keys.shiftKey || this.keys.ctrlKey || this.keys.altKey) {
      this.repaintNeighbors (node);
      } else {
         this.memory.select.toggle (node);
      }
      this.lastClick = performance.now();   //TODO oldClick / newClick
   }//end Fct


   //------------------------Start onNodeHover------------------------------------------//
   onNodeHover (node, prevNode) {
      try {
         this.memory.hover.show (node);
         if (node) this.animation.addHalo (node.__threeObj, "Hover", 0xa1a3fc);
         if (prevNode) this.animation.removeHalo (prevNode.__threeObj, "Hover");
      } catch (e) { console.error (e); }
   }//end Fct


   //------------------------Start onLinkClick------------------------------------------//
   onLinkClick (link) {
   }//end Fct


   //------------------------Start onLinkHover------------------------------------------//
   onLinkHover (link, prevLink) {
   }//end Fct


   //------------------------Start onClick----------------------------------------------//
   onClick (event) {
      this.keys = event;
   }//end Fct


   //------------------------Start onMouseDown------------------------------------------//
   onMouseDown (event) {
      if (event.button == 0 && !event.altKey && event.shiftKey != event.ctrlKey && this.graph.enablePointerInteraction () ) {
         //console.log (event)
         event.onMouseUp = this.onMouseUp.bind (this);
         this.node.addEventListener ("mouseup", event.onMouseUp, true);
         event.onMouseMove = this.onMouseMove.bind (this);
         this.node.addEventListener ("mousemove", event.onMouseMove, true);
         this.bbSelect = event;
         this.bbSelect.moved = false;
         event.stopPropagation ();
      }
   }//end Fct


   //------------------------Start onMouseMove------------------------------------------//
   onMouseMove (event) {
      //console.log (event)

      if (this.bbSelect.moved == true || Math.abs (this.bbSelect.clientX - event.clientX) > 5 && Math.abs (this.bbSelect.clientY - event.clientY) > 5) {
         this.bbSelect.moved = true;

         var rect = [this.bbSelect.layerX, this.bbSelect.layerY, event.layerX, event.layerY];
         this.bbSelect.bb = this.animation.addBoundingBox (this.bbSelect.bb, rect);
      }
   }//end Fct


   //------------------------Start onMouseUp--------------------------------------------//
   onMouseUp (event) {
      if (this.bbSelect.moved) {
         //console.log (event)
         this.animation.addBoundingBox (this.bbSelect.bb);
         this.bbTimeStamp = event.timeStamp;   // Effectively prevent onNodeClick

         var rect = [
            Math.min (this.bbSelect.layerX, event.layerX),
            Math.min (this.bbSelect.layerY, event.layerY),
            Math.max (this.bbSelect.layerX, event.layerX),
            Math.max (this.bbSelect.layerY, event.layerY)
         ];
         var nodelist = [];

         this.data.nodes.forEach (node => {
            var pos = this.animation.getScreenPosition (node.__threeObj);
            if (pos.x >= rect[0] && pos.x <= rect[2] && pos.y >= rect[1] && pos.y <= rect[3])
               nodelist.push (node);
         });//end for

         if (this.bbSelect.shiftKey)
            this.activateNeighbors (nodelist);
         else
            nodelist.forEach (node => { if (node.__threeObj.visible) this.memory.select.toggle (node) });

         event.stopPropagation ();
      }
      this.node.removeEventListener ("mouseup", this.bbSelect.onMouseUp, true);
      event.onMouseMove = this.onMouseMove.bind (this);
      this.node.removeEventListener ("mousemove", this.bbSelect.onMouseMove, true);
      this.bbSelect = undefined;
   }//end Fct


   //------------------------Start resize-----------------------------------------------//
   resize (height, width) {
      if (height >= 0) this.graph.height(height);
      if (width >= 0) this.graph.width(width);
   }//end Fct


   //------------------------Start onResize---------------------------------------------//
   onResize (node) {
      if (this.graph === undefined) return;
      if (node == undefined) node = this.node;
      this.graph.height(parseInt(window.getComputedStyle(node)['height'].split('px')[0]));
      this.graph.width(parseInt(window.getComputedStyle(node)['width'].split('px')[0]));
   }//end Fct


   //------------------------Start registerResize---------------------------------------//
   registerResize () {
      this.mutation = new MutationObserver(mutations => {
         mutations.forEach(mutation => {
            this.onResize(mutation.target);
         });
      });

      this.mutation.observe(this.node, { attributes: true, childList: false, characterData: false, attributeFilter: ["style"] });
      this.mutation.observe(this.memory.node, { attributes: true, childList: false, characterData: false, attributeFilter: ["style"] });
   }//end Fct


   //------------------------Start addNodeLabel-----------------------------------------//
   addNodeLabel (node) {
      switch (node.type) {
      //case "AgeingFactor": this.animation.addFrontText (node.name, {color: this.colors[node.type + "_" + node.subtype], size: 4, height: 1, node: node}); break;
      case "AlleleType": this.animation.addFrontText (node.id, {color: node.__threeObj.material.color, size: 2, height: 1, node: node}); break;
      case "GeneOntologyTerm": this.animation.addFrontText (node.term, {color: node.__threeObj.material.color, size: 2, height: 1, node: node}); break;
      case "Observation": {
         if (node.subtype == "lifespan") {
            var name = this.memory.menu.na;
            if (node.lifespanChange && node.lifespanEffectSignificant != "none") {
               name = parseFloat (node.lifespanChange);
               if (name >= 0) name = "+" + name.toFixed (0) + "%";
               else name = name.toFixed (0) + "%";
            }
            this.animation.addFrontText (name, {color: node.__threeObj.material.color, bgcolor: false, size: 2, height: 1, node: node});
         }
      } break;
      case "Species": {
         var text = node.id.split (" ");
         text[0] = text[0][0] + '.';
         text = text.join (" ");
         this.animation.addFrontText (text, {color: node.__threeObj.material.color, bgcolor: 0x808080, size: 2, height: 1, node: node});
      } break;
      default: this.animation.addFrontText (node.name, {color: node.__threeObj.material.color, size: 2, height: 1, node: node});
      };//end switch
   }//end Fct


   //------------------------Start removeNodeLabel--------------------------------------//
   removeNodeLabel (node) {
      this.animation.removeFrontText (node.__threeLabel);
   }//end Fct


   //------------------------Start resizeNodeLabel--------------------------------------//
   resizeNodeLabel (node, factor) {
      this.animation.resizeFrontText (node.__threeLabel, factor);
   }//end Fct


   //------------------------Start resizeNode-------------------------------------------//
   resizeNode (node, factor) {
      this.animation.resizeObj (node.__threeObj, factor);
      if (factor) node._scale = node.__threeObj.scale.clone ();
      else delete node._scale;
   }//end Fct


   //------------------------Start toggleVisibility-------------------------------------//
   toggleVisibility (type) {
      var { nodes, links } = this.data;
      this.active[type] = !this.active[type];
      var opacity = this.active[type] ? 1 : 0;

      if (type == "GOI") {
         this.data.nodes.forEach (node => {
            if (!node._goi) return;

            if (this.active.GOI)
               this.animation.addHalo (node.__threeObj, "GOI", this.colors.AgeingFactor_gene);
            else
               this.animation.removeHalo (node.__threeObj, "GOI");
         });
      } else {
         nodes.forEach (node => {
            if (node.type != type) return;

            // node.__threeObj.material.opacity = opacity;
            // node.links.forEach (e => e.__lineObj.material.opacity = opacity);
            node.__threeObj.visible = opacity ? true : false;
            node.links.forEach (e => e.__lineObj.visible = opacity ? true : false);

            if (node.__threeLabel) node.__threeLabel.visible = opacity ? true : false;
         });//end for
      }
   }//end Fct


   //------------------------Start highlightNeighbors-----------------------------------//
   // highlightNeighbors (node) {
   //    if (!node || node == this.selected) {
   //       this.selected = undefined; this.graph.nodeRelSize(1)
   //       this.graph.nodeColor('color')
   //       this.graph.linkColor('color')
   //       this.update();
   //    } else {
   //       this.data.nodes.forEach(e => {
   //          //e.highlight = 'rgba (0,0,0,0.1)';
   //          e.highlight = e.color + "00";
   //       });
   //       node.neighbors.forEach(e => {
   //          e.highlight = e.color;
   //       });

   //       this.data.links.forEach(e => {
   //          //e.highlight = 'rgba (0,0,0,0.1)';
   //          e.highlight = e.color + "00";
   //       });
   //       node.links.forEach(e => {
   //          e.highlight = e.color;
   //       });

   //       node.highlight = node.color;
   //       this.selected = node;

   //       this.graph.nodeColor('highlight');
   //       this.graph.linkColor('highlight');
   //       this.update();
   //    }
   // }//end Fct


   //------------------------Start updateNeighbors--------------------------------------//
   // updateNeighbors (node) {
   //    if (!node) {   // No node is selected --> everythin is visible
   //       this.graph.nodeThreeObject(undefined)
   //       this.data = this.allData;
   //       this.selected = undefined;
   //    } else if (!this.selected) {   // Previously nothing selected --> subgraph
   //       //this.graph.nodeThreeObject (node => new THREE.Mesh (node.__threeObj.geometry, new THREE.MeshBasicMaterial({ color: node.color }) ) )
   //       // Restore current camera position
   //       console.log (this.camera)
   //       setTimeout ( (position) => this.graph.cameraPosition (position), 1000, this.camera.position.clone () );
   //       this.data = { nodes: [node, ...node.neighbors], links: node.links };
   //       this.graph.cooldownTicks (0);
   //       this.selected = node;
   //    } else if (node.id == this.selected.id && this.keys.shiftKey) {
   //       var data = { nodes: [node, ...node.neighbors], links: node.links };
   //       node.neighbors.forEach(node => {
   //          data.nodes = [...data.nodes, ...node.neighbors];
   //          data.links = [...data.links, ...node.links];
   //       })
   //       //this.graph.nodeThreeObject (node => new THREE.Mesh (node.__threeObj.geometry, new THREE.MeshBasicMaterial({ color: node.color }) ) )
   //       this.data = data;
   //       this.graph.cooldownTicks(0);
   //       this.selected = node;
   //    } else if (node.id == this.selected.id) {
   //       this.graph.nodeThreeObject(undefined)
   //       this.data = this.allData;
   //       this.selected = undefined;
   //    } else {
   //       //this.graph.nodeThreeObject (node => new THREE.Mesh (node.__threeObj.geometry, new THREE.MeshBasicMaterial({ color: node.color }) ) )
   //       this.data = { nodes: [node, ...node.neighbors], links: node.links };
   //       this.graph.cooldownTicks(0);
   //       this.selected = node;
   //    }
   // }//end Fct


   //------------------------Start neighborLSChange-------------------------------------//
   neighborLSChange () {
      this.data.nodes.forEach (node => {
         if (node.type != "AgeingFactor") return;

         var inc = 0, dec = 0, none = 0, count = 0;

         node.neighbors.forEach (ob => {
            if (ob.type != "Observation" || ob.subtype != "lifespan") return;

            if (ob.lifespanEffectSignificant == "increased") ++inc;
            else if (ob.lifespanEffectSignificant == "decreased") ++dec;
            else ++none;
            ++count;
         });//end for 2

         var mode;

         if (count == 0) return;
         if (dec == count) mode = ['or', 0xff0000, 1];
         else if (inc == count) mode = ['og', 0x00ff00, 1];
         else if (none == count) mode = ['ob', 0x808080, 1];
         else if (dec / count >= 0.8) mode = ['tr', 0xff0000, 0.6];
         else if (inc / count >= 0.8) mode = ['tg', 0x00ff00, 0.6];
         else if (none / count >= 0.8) mode = ['tb', 0xff0000, 0.6];
         else mode = ['o', 0xffa500, 1];

         node.LOtype = mode[0];
         node.__threeObj.material = node.__threeObj.material.clone ();
         node.__threeObj.material.color.set (mode[1]);
         node.__threeObj.material.opacity = mode[2];
      });//end for 1
   }//end Fct


   //------------------------Start resizeByLO-------------------------------------------//
   resizeByLO (nodelist, factor) {
      if (!nodelist) nodelist = this.data.nodes;
      if (!factor) {
         nodelist.forEach (node => {
            if (node.lifespanChange !== undefined && node.lifespanEffectSignificant != "none")
               this.memory.graph.resizeNode (node, "reset")
         });//end for
         return;
      }

      var min_ = 100000, max_ = 0;
      nodelist.forEach (node => { if (node.lifespanChange) {min_ = Math.min (min_, Math.abs (node.lifespanChange) ); max_ = Math.max (max_, Math.abs (node.lifespanChange) ); } });
      var scale = max_ - min_;
      var sizeNode = (node) => (Math.abs ( (node.lifespanChange - min_) / scale) - (Math.abs(factor) * 0.4) ) * factor;

      nodelist.forEach (node => {
         if (node.lifespanChange !== undefined && node.lifespanEffectSignificant != "none")
            this.memory.graph.resizeNode (node, sizeNode (node) )
      });//end for
   }//end Fct


   //------------------------Start repaintNeighbors-------------------------------------//
   repaintNeighbors (node) {
      if (!node) {   // No node is selected --> everythin is visible
         this.activateNeighbors ();
         //this.graph.nodeThreeObject(undefined);
         this.selected = undefined;
      } else if (this.keys.shiftKey) {   // Extend
         var nodelist = [node, ...node.neighbors];
         node.neighbors.forEach (node => { nodelist = [...nodelist, ...node.neighbors]; });
         this.activateNeighbors (nodelist);
         //nodelist.forEach (node => { node.__threeObj.geometry = new THREE.DodecahedronGeometry (5) });
         this.selected = node;
      } else if (!this.selected) {   // Previously nothing selected --> subgraph
         var nodelist = [node, ...node.neighbors];
         //nodelist.forEach (node => { node.__threeObj.geometry = new THREE.DodecahedronGeometry (5) });
         this.activateNeighbors (nodelist);
         this.selected = node;
      // } else if (node.id == this.selected.id && this.keys.shiftKey) {   // Extend
      //    var nodelist = [node, ...node.neighbors];
      //    node.neighbors.forEach (node => { nodelist = [...nodelist, ...node.neighbors]; });
      //    this.activateNeighbors (nodelist);
      //    this.selected = node;
      } else if (node.id == this.selected.id) {
         //this.graph.nodeThreeObject(undefined);
         this.activateNeighbors ();
         this.selected = undefined;
      } else {
         var nodelist = [node, ...node.neighbors];
         //nodelist.forEach (node => { node.__threeObj.geometry = new THREE.DodecahedronGeometry (5) });
         this.activateNeighbors (nodelist);
         this.selected = node;
      }
   }//end Fct


   //------------------------Start activateNeighbors------------------------------------//
   activateNeighbors (nodelist) {
      var { nodes, links } = this.data;

      if (!nodelist || !nodelist.length) {
         nodes.forEach (e => { e.__threeObj.visible = this.active[e.type]; if(e.__threeLabel) e.__threeLabel.visible = this.active[e.type] });
         links.forEach (e => { e.__lineObj.visible = this.active[e.source.type] && this.active[e.target.type] });
         return;
      }

      nodes.forEach (e => { e.__threeObj.visible = false; if(e.__threeLabel) e.__threeLabel.visible = false });
      links.forEach (e => { e.__lineObj.visible = false });

      nodelist.forEach (e => { e.__threeObj.visible = this.active[e.type]; if(e.__threeLabel) e.__threeLabel.visible = this.active[e.type] });


      nodelist.forEach (node => {
         node.links.forEach (link => {
            if (nodelist.includes (link.source) && nodelist.includes (link.target) )
               link.__lineObj.visible = this.active[link.source.type] && this.active[link.target.type];
         })
      });//end for
   }//end Fct


   //------------------------Start selectNeighbors--------------------------------------//
   // selectNeighbors () {
   //    //Toggle stores whether the highlighting is on
   //    var toggle = 0;
   //    //Create an array logging what is connected to what
   //    var linkedByIndex = {};
   //    for (i = 0; i < graph.nodes.length; i++) {
   //       linkedByIndex[i + "," + i] = 1;
   //    };
   //    graph.links.forEach(function (d) {
   //       linkedByIndex[d.source.index + "," + d.target.index] = 1;
   //    });
   //    //This function looks up whether a pair are neighbours
   //    function neighboring(a, b) {
   //       return linkedByIndex[a.index + "," + b.index];
   //    }
   //    function connectedNodes() {
   //       if (toggle == 0) {
   //          //Reduce the opacity of all but the neighbouring nodes
   //          d = d3.select(this).node().__data__;
   //          node.style("opacity", function (o) {
   //             return neighboring(d, o) | neighboring(o, d) ? 1 : 0.1;
   //          });
   //          link.style("opacity", function (o) {
   //             return d.index == o.source.index | d.index == o.target.index ? 1 : 0.1;
   //          });
   //          //Reduce the op
   //          toggle = 1;
   //       } else {
   //          //Put them back to opacity=1
   //          node.style("opacity", 1);
   //          link.style("opacity", 1);
   //          toggle = 0;
   //       }
   //    }
   // }//end Fct


   //------------------------Start export-----------------------------------------------//
   // https://stackoverflow.com/questions/923885/capture-html-canvas-as-gif-jpg-png-pdf
   export (format, size) {
      var origsize = [this.canvas.height, this.canvas.width];

      function save (blob, filename) {
         var anchor = document.createElement ('a');
         anchor.download = filename;
         anchor.href = (window.webkitURL || window.URL).createObjectURL (blob);
         anchor.dataset.downloadurl = [blob.type, anchor.download, anchor.href].join (':');
         anchor.style.setProperty ("max-height", "0px");
         //anchor.innerHTML = "JANet Export";
         document.body.appendChild (anchor);
         anchor.click ();
         document.body.removeChild (anchor);
      };


      switch (format) {
         case "PNG": {
            try {
               if (size) {
                  this.resize (size[0], size[1]);
                  if (origsize[1] / origsize[0] > size[1] / size[0])
                     this.camera.userData = { height: () => origsize[1] * size[0] / size[1], width: () => origsize[1] };
                  else
                     this.camera.userData = { height: () => origsize[0], width: () => origsize[0] * size[1] / size[0] };
               }

               requestAnimationFrame ( () => {
                  requestAnimationFrame (timestamp => {   // A second frame for the resizing thing
                     var image = this.canvas.toBlob (blob => {
                        save (blob, "janet_" + (new Date ().toISOString () ) + ".png");
                     }, "image/png");
                     this.resize (origsize[0], origsize[1]);
                     this.camera.userData = this.graph;
                  });//end Lambda 2
               });//end Lambda 1
            } catch (e) {
               this.resize (origsize[0], origsize[1]);
               this.camera.userData = this.graph;
            }
         } break;
         case "GLTF": {
            try {
               requestAnimationFrame (timestamp => {   // A second frame for the resizing thing
                  // Instantiate a exporter
                  var exporter = new THREE.GLTFExporter();

                  // Parse the input and generate the glTF output
                  exporter.parse (scene, function (gltf) {
                     gltf = JSON.stringify (gltf, null, 2);
                     var blob = new Blob ([gltf], {type: "application/json"});
                     save (blob, "janet_" + (new Date ().toISOString () ) + ".gltf");
                              });
               });//end Lambda 1
            } catch (e) {}
         } break;
         case "Blender": {
            try {
               requestAnimationFrame (timestamp => {   // A second frame for the resizing thing
                  blender = JSON.stringify (this.scene.toJSON (), null, 2);
                  var blob = new Blob ([blender], {type: "application/json"});
                  save (blob, "janet_" + (new Date ().toISOString () ) + ".gltf");
               });//end Lambda 1
            } catch (e) {}
         } break;
         case "PDF": {
            requestAnimationFrame (timestamp => {
               var imgData = this.canvas.toDataURL ("image/png");
               var doc = new jsPDF ('p', 'mm');
               doc.addImage (imgData, 'PNG', 10, 10);
               doc.save ("janet_" + (new Date ().toISOString () ) + ".pdf");
            });

         } break;
         case "XGMML": {
            this.memory.statusbar.content = "Warning: XGMML not fully supported.";

            // var doc = export_XGMML (this.allData.nodes, this.allData.links);
            var doc = export_XGMML (this.data.nodes, this.data.links);

            var xml = new XMLSerializer ().serializeToString (doc);
            var blob = new Blob ([xml], {type: doc.contentType});
            save (blob, "janet_" + (new Date ().toISOString () ) + ".xml");
         } break;
      };//end switch
   }//end Fct
};//end Class
