
//---------------------------Class Request----------------------------------------------//
class Request {
   constructor (api, options, settings) {
      this.server = window.JANET_SERVER;
      this.viewer = window.JANET_VIEWER || this.server + "Viewer/";

      this.api = api;
      this.options = options;

      if (settings === undefined) settings = {};

      this.mode = settings.mode || "API";
      this.method = settings.method || "GET";
      this.user = settings.user;
      this.password = settings.password;
      this.data = settings.data;
      this.contentType = settings.contentType;
      this.convert = settings.convert || function (data) { if (data.json) return data.json (); else return JSON.parse (data) };

      this.onerror = console.error;
   }//end Constructor


   //------------------------Start toString---------------------------------------------//
   toString () {
      if (this.mode == "API")
         return this.server + "api/" + this.api + "?" + this.options;
      else
         return this.viewer + this.api + "?" + this.options;
   }//end Fct


   //------------------------Getter options---------------------------------------------//
   get options () {
      var opts = "";

      if (typeof (this.opts) == "string") {
         opts = this.opts;
         if (opts[0] == '?') opts = opts.substr (1);

      } else {
         opts = [];
         for (var key in this.opts) {
            if (this.opts.hasOwnProperty (key) )
               opts.push (key + "=" + this.opts[key]);
         }//end for
         opts = opts.join ('&');
      }

      return opts;
   }//end Fct


   //------------------------Setter options---------------------------------------------//
   set options (opts) {
      if (!this.opts) this.opts = {};
      if (!opts) return;

      if (typeof (opts) == "string") {
         if (opts[0] == '?') opts = opts.substr (1);

         var o = opts.split ("&");

         o.forEach (e => {
            var key = e.split ("=", 1)[0];
            var value = e.substr (key.length + 1);
            this.opts[key] = value;
         });//end for

      } else {
         Object.keys (opts).forEach (key => this.opts[key] = opts[key] );
      }
   }//end Fct


   //------------------------Start createRequestObj-------------------------------------//
   createRequestObj () {
      if (this.api != "search") return this;
      if (typeof (opts) == "string") return this;
      var options = this.options;   //FIXME inefficient
      if (options.length < 32000) return this;

      var opts = [];
      for (var key in this.opts) {
         if (this.opts.hasOwnProperty (key) )
            opts.push (key + "=" + this.opts[key]);
      }//end for

      return {
         method: 'POST',
         toString: () => {
            if (this.mode == "API")
               return this.server + "api/" + this.api + "?body=true";
            else
               return this.viewer + this.api + "?body=true";
         },
         data: opts.join ('\n')
      };
   }//end Fct


   //------------------------Start catch------------------------------------------------//
   catch (callback) {
      this.onerror = callback;
      return this;
   }//end Fct


   //------------------------Start load-------------------------------------------------//
   // Dynamically load json
   // https://codepen.io/KryptoniteDove/post/load-json-file-locally-using-pure-javascript
   load (callback) {
      var xobj = new XMLHttpRequest ();
      var req = this.createRequestObj ();

      xobj.open (req.method, req, false, this.user, this.password);
      xobj.onerror = this.onerror;
      xobj.send (req.data);

      var response = this.convert (xobj.responseText);

      if (callback) return callback (response);
      else return response;
   }//end Fct


   //------------------------Start async------------------------------------------------//
   // Dynamically load json
   // https://codepen.io/KryptoniteDove/post/load-json-file-locally-using-pure-javascript
   async (callback) {
      var request = new Promise ( (resolve, reject) => {
         var xobj = new XMLHttpRequest ();
         var req = this.createRequestObj ();

         //xobj.overrideMimeType ("application/json");
         xobj.open (req.method, req, true, this.user, this.password);

         xobj.onerror = (event) => { reject (event); };
         xobj.onload = (event) => {
            try {
               var obj = this.convert (xobj.responseText);
               if (obj === undefined) throw "unknown error";
               else if (obj.error) throw obj.error;
               else resolve (obj);
            } catch (ex) {
               reject (ex);
            }//end catch
         };
         // xobj.onreadystatechange = () => {
         //    if (xobj.readyState == 4 && xobj.status == "200") {
         //       resolve (xobj.response);
         //    }
         // };

         xobj.send (req.data);
      }).catch (this.onerror);

      if (callback) return request.then (callback);
      else return request;
   }//end Fct


   //------------------------Start fetch------------------------------------------------//
   // Dynamically load json
   fetch (callback) {
      var request = fetch (this).catch (this.onerror).then (this.convert).catch (this.onerror);

      if (callback) return request.then (callback).catch (this.onerror);
      else return request;
   }//end Fct

};//end Class



//---------------------------Start RequestSearch----------------------------------------//
function RequestSearch (opts, settings) {
   return new Request ("search", opts, settings);
}//end Fct



//---------------------------Start RequestSpecial---------------------------------------//
function RequestSpecial (opts, settings) {
   return new Request ("special", opts, settings);
}//end Fct



//---------------------------Start RequestList------------------------------------------//
function RequestList (opts, settings) {
   if (opts.type == "GOI" && settings === undefined) settings = { method: 'POST' };
   if (opts.type == "GOI" && settings.method === undefined) settings.method = 'POST';

   return new Request ("list", opts, settings);
}//end Fct



//---------------------------Start RequestList------------------------------------------//
function RequestV3rd (url, opts, settings) {
   settings = settings || {};
   settings.mode = "VIEWER";
   settings.convert = function (data) { return data };
   return new Request ("3rdParty/" + url, opts, settings);
}//end Fct
