
//---------------------------Class Select-----------------------------------------------//
class Select {
   // Restore constructor
   constructor (memory, node) {
      this.memory = memory;
      this.node = node;
      this.countCache = [0, 0];
      this.valid = true;
   }//end Constructor


   //------------------------Getter selected--------------------------------------------//
   get selected () {
      var nodelist = [];
      this.memory.graph.data.nodes.forEach (node => {
         if (node._selected == true)
            nodelist.push (node);
      });//end for
      return nodelist;
   }//end Fct


   //------------------------Setter selected--------------------------------------------//
   set selected (nodelist) {
      this.memory.graph.data.nodes.forEach (node => {
         this.remove (node);
      });//end for
      nodelist.forEach (node => this.add (node) );
      return nodelist;
   }//end Fct


   //------------------------Getter count-----------------------------------------------//
   get count () {
      if (this.valid) return this.countCache;
      this.valid = true;
      var selected = this.selected;
      this.countCache = [selected.length, 0];
      var count = 0;
      this.graph.data.links.firEach (link => {
         if (link.source._selected && link.target._selected) ++count;
      });//end for
      this.countCache[1] = count;
      return this.countCache;
   }//end Fct


   //------------------------Getter marked----------------------------------------------//
   // Alias: visualized
   get marked () {
      var nodelist = [];
      this.memory.graph.data.nodes.forEach (node => {
         if (node.__threeObj.visible == true)
            nodelist.push (node);
      });//end for
      return nodelist;
   }//end Fct


   //------------------------Setter marked----------------------------------------------//
   // Alias: visualized
   set marked (nodelist) {
      this.memory.graph.activateNeighbors (nodelist);
      return nodelist;
   }//end Fct


   //------------------------Start toggle-----------------------------------------------//
   toggle (node) {
      this.valid = false;
      node._selected = !node._selected;
      if (node._selected) {
         this.memory.graph.animation.addHalo (node.__threeObj, "Selected");
      } else {
         this.memory.graph.animation.removeHalo (node.__threeObj, "Selected");
      }
   }//end Fct


   toggleNodes (nodelist) {
      nodelist.forEach (node => this.toggle (node) );
   }//end Fct


   //------------------------Start add--------------------------------------------------//
   add (node) {
      this.valid = false;
      node._selected = true;
      this.memory.graph.animation.addHalo (node.__threeObj, "Selected");
   }//end Fct


   addNodes (nodelist) {
      nodelist.forEach (node => this.add (node) );
   }//end Fct


   //------------------------Start remove-----------------------------------------------//
   remove (node) {
      this.valid = false;
      node._selected = false;
      this.memory.graph.animation.removeHalo (node.__threeObj, "Selected");
   }//end Fct


   removeNodes (nodelist) {
      nodelist.forEach (node => this.remove (node) );
   }//end Fct


   //------------------------Start merge------------------------------------------------//
   merge (list) {
      if (!list || list.length <= 0) return [];
      if (list.length == 1) return list[0];

      var a = list[0];

      list.forEach ( (e,p) => {
         if (p == 0) return;
         e.forEach (e => {
            var i = a.indexOf (e);
            if (i < 0) a.push (e);
         });//end for 2
      });//end for 1

      return a;
   }//end Fct


   //------------------------Start exclude----------------------------------------------//
   exclude (list) {
      if (!list || list.length <= 0) return [];
      if (list.length == 1) return list[0];

      var a = list[0];

      list.forEach ( (e,p) => {
         if (p == 0) return;
         e.forEach (e => {
            var i = a.indexOf (e);
            if (i >= 0) a.splice (i, 1);
         });//end for 2
      });//end for 1

      return a;
   }//end Fct


   //------------------------Start search-----------------------------------------------//
   search (str) {
      // var op = (a) => a == str;
      // if (str.search (/^ *-?[0-9.]+ *$/) == 0) {
      //    str = parseFloat (str);
      // } else if (str.search (/^ *> *-?[0-9.]+ *$/) == 0) {
      //    str = parseFloat (str.substr(1) );
      //    op = (a) => a > str;
      // } else if (str.search (/^ *>= *-?[0-9.]+ *$/) == 0) {
      //    str = parseFloat (str.substr(2) );
      //    op = (a) => a >= str;
      // } else if (str.search (/^ *< *-?[0-9.]+ *$/) == 0) {
      //    str = parseFloat (str.substr(1) );
      //    op = (a) => a < str;
      // } else if (str.search (/^ *<= *-?[0-9.]+ *$/) == 0) {
      //    str = parseFloat (str.substr(2) );
      //    op = (a) => a <= str;
      // } else throw "search string not valid";

      var pos = str.search (" ");
      var type = str.substr (0, pos);
      var value = str.substr (pos + 1);

      return this.searchType ("property", type, value);
   }//end Fct


   //------------------------Start searchType-------------------------------------------//
   searchType (type, op, str) {
      var nodelist = [];
      if (str == "true") str = true;
      else if (str == "false") str = false;

      switch (type) {
      case "name": {
         this.memory.graph.data.nodes.forEach (node => {
            if ( (node.name && node.name.search (str) >= 0) ||
                 (node.synonyms && node.synonyms.some (e => e.search (str) >= 0) ) )
               nodelist.push (node);
         });//end for
      } break;
      case "type": {
         this.memory.graph.data.nodes.forEach (node => {
            if (node.type == op)
               nodelist.push (node);
         });//end for
      } break;
      case "degree": {
         op = eval ("(a) => a " + op + " str");
         str = parseFloat (str);

         this.memory.graph.data.nodes.forEach (node => {
            if (node.links && op (node.links.length) )
               nodelist.push (node);
         });//end for
      } break;
      case "property": {
         this.memory.graph.data.nodes.forEach (node => {
            if (node[op]) {
               console.log (node[op])
               console.log (str)
               if (node[op] instanceof Array) {
                  if (node[op].some (e => e.search (str) >= 0) )
                     nodelist.push (node);
               } else if (node[op] === str) {
                  nodelist.push (node);
               } else if (node[op].search (str) >= 0) {
                  nodelist.push (node);
               }
            }
         });//end for
      } break;
      case "goi": {
         this.memory.graph.data.nodes.forEach (node => {
            if (node._goi)
               nodelist.push (node);
         });//end for
      } break;
      case "goi_ho_base": {
         this.memory.graph.data.nodes.forEach (node => {
            if (node._req && !node._goi)
               nodelist.push (node);
         });//end for
      } break;
      case "lifespan": {
         op = eval ("(a) => a " + op + " str");
         str = parseFloat (str);

         this.memory.graph.data.nodes.forEach (node => {
            if (node.lifespanChange && op (node.lifespanChange) )
               nodelist.push (node);
         });//end for
      } break;
      case "labeled": {
         this.memory.graph.data.nodes.forEach (node => {
            if (node.__threeLabel)
               nodelist.push (node);
         });//end for
      } break;
      case "neighbors": {
         this.resolveSubNetworks (this.selected, 1).forEach (nodes => {
            if (!op)
               nodelist = [...nodelist, ...nodes];
            else
               nodes.forEach (node => {
                  if (node.type == op)
                     nodelist.push (node);
               });//end for 2
         });//end for 1
      } break;
      case "subnetwork": {
         this.resolveAllSubNetworks (this.selected).forEach (nodes => {
            nodelist = [...nodelist, ...nodes];
         });//end for
      } break;
      case "invert": {
         this.memory.graph.data.nodes.forEach (node => {
            if (!node._selected && (!op || node.type == op) )
               nodelist.push (node);
         });
      } break;
      default: throw "search type unknown";
      };//end switch

      return nodelist;
   }//end Fct


   //------------------------Start resolveSubNetworks-----------------------------------//
   resolveSubNetworks (nodelist, depth) {
      if (!nodelist) return [];

      var nodeslist = [];

      nodelist.forEach (node => nodeslist.push (this.resolveSubNetwork (node, depth) ) );

      return nodeslist;
   }//end Fct


   //------------------------Start resolveSubNetwork------------------------------------//
   resolveSubNetwork (node, depth = 99999) {
      var resolve = (nodes, node) => {
         var list = [];
         node.neighbors.forEach (n => {
            if (!n._deleted && nodes.indexOf (n) == -1) {
               list.push (n);
               nodes.push (n)
            }
         });//end for
         return list;
      };


      var nodes = [];
      var list = [node];
      do {
         var list2 = [];
         list.forEach (node => list2 = [...list2, ...resolve (nodes, node)]);
         list = list2;
      } while (list.length > 0 && --depth > 0);

      return nodes;
   }//end Fct


   //------------------------Start resolveAllSubNetworks--------------------------------//
   resolveAllSubNetworks (nodelist, depth = 99999) {
      if (!nodelist) nodelist = this.memory.graph.data.nodes;

      var resolve = (allnodes, nodes, node) => {
         var list = [];
         node.neighbors.forEach (n => {
            if (!n._deleted && nodes.indexOf (n) == -1) {
               list.push (n);
               nodes.push (n)
               allnodes.push (n)
            }
         });//end for
         return list;
      };


      var allnodes = [];
      var nodeslist = [];
      nodelist.forEach (node => {
         if (allnodes.indexOf (node) != -1) return;
         var nodes = [];
         var list = [node];
         var depth_ = depth;
         do {
            var list2 = [];
            list.forEach (node => list2 = [...list2, ...resolve (allnodes, nodes, node)]);
            list = list2;
         } while (list.length > 0 && --depth_ > 0);
         nodeslist.push (nodes);
      });//end for 1

      return nodeslist;
   }//end Fct
};//end Class
