"""JANet REST API
"""

import os, sys
import io
import re

from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt

import listNodes_utils

BASE_DIR = os.path.dirname (os.path.dirname (os.path.abspath (__file__) ) )
sys.path.insert (0, os.path.join (BASE_DIR, "Tools") )

from Arguments import Arguments
from Neo4j import Neo4j


neo4j = Neo4j ()



##---------------------------Help Message------------------------------------------##
_help = """\
TODO: write this help page ;p
"""



##---------------------------Start listAgeingFactor--------------------------------##
def listAgeingFactor (request, response):
   response['nodes'] = nodes = []

   for record in neo4j.eval ('match (af:AgeingFactor) return distinct af.id as id', {}):
      nodes.append (dict (record.items () ) )

   return response



##---------------------------Start listAlleleType----------------------------------##
def listAlleleType (request, response):
   response['nodes'] = nodes = []

   for record in neo4j.eval ('match (at:AlleleType) return distinct at.id as id', {}):
      nodes.append (dict (record.items () ) )

   return response



##---------------------------Start listCitation------------------------------------##
def listCitation (request, response):
   response['nodes'] = nodes = []

   for record in neo4j.eval ('match (ct:Citation) return distinct ct.id as id', {}):
      nodes.append (dict (record.items () ) )

   return response



##---------------------------Start listSpecies-------------------------------------##
def listSpecies (request, response):
   response['nodes'] = nodes = []

   for record in neo4j.eval ('match (sp:Species) return distinct sp.id as id', {}):
      nodes.append (dict (record.items () ) )

   return response



##---------------------------Start startup-----------------------------------------##
def GenesOfInterest (request, response):
   #response['error'] = "not implemented"

   if request.method != 'POST':
      response['error'] = "Request must be a HTTP POST"
      return response

   # Check input
   if len (request.body) == 0:
      response['error'] = "missing list of GOIs"
      return response

   content = request.body.decode ('ascii')

   if not re.match ('^[a-zA-Z0-9\'.\\n\t ,_-]+$', content):
      response['error'] = "list of GOIs is not correct"
      return response


   lines = io.StringIO (content)

   GOIs = []
   AF_names = []
   AF_ncbis = []


   for line in lines:
      if len (line) == 0: continue
      if line[-1] == '\n': line = line[:-1]

      if '\t' in line:
         AF_name, SP_id = line.split ("\t", 1)
         if AF_name.strip() in AF_names: continue
         GOIs.append ({'AF_name': AF_name.strip(), 'SP_id': SP_id.strip()})
         AF_names.append (AF_name.strip() )
      elif ',' in line:
         AF_name, SP_id = line.split (",", 1)
         if AF_name.strip() in AF_names: continue
         GOIs.append ({'AF_name': AF_name.strip(), 'SP_id': SP_id.strip()})
         AF_names.append (AF_name.strip() )
      elif ' ' in line:
         AF_name, SP_id = line.split (" ", 1)
         if AF_name.strip() in AF_names: continue
         GOIs.append ({'AF_name': AF_name.strip(), 'SP_id': SP_id.strip()})
         AF_names.append (AF_name.strip() )
      else:
         if line.strip() in AF_ncbis: continue
         GOIs.append ({'AF_ncbi_gene_id': line.strip()})
         AF_ncbis.append (line.strip() )


   return listNodes_utils.resolveGenesOfInterest (response, GOIs)



##---------------------------Start entry-------------------------------------------##
@csrf_exempt
def entry (request):
   response = {}

   type_ = request.GET.get ("type")

   if "help" in request.GET:
      response["help"] = _help

   elif not type_:
      response['error'] = "No command defined; use api/search?help"

   elif type_ == "AgeingFactor":
      listAgeingFactor (request, response)

   elif type_ == "AlleleType":
      listAlleleType (request, response)

   elif type_ == "Citation":
      listCitation (request, response)

   elif type_ == "Species":
      listSpecies (request, response)

   elif type_ == "GenesOfInterest" or type_ == 'GOI':
      GenesOfInterest (request, response)

   else:
      response['error'] = "Command not found"

   return JsonResponse (response)
