"""JANet REST API
"""

import os, sys
import io
import re

from django.http import JsonResponse, HttpResponse
from django.views.decorators.csrf import csrf_exempt


import search_utils



##---------------------------Help Message------------------------------------------##
_help = """\
TODO: write this help page ;p
"""



##---------------------------Start searchAllAgeingFactors--------------------------##
def searchAllAgeingFactors (request, response):
   AF_type         = request.GET.get ("subtype")
   AF_species      = request.GET.get ("species")
   annotationNodes = request.GET.get ("nodes", "")
   layout          = request.GET.get ("layout", "")


   # Check input
   if not re.match ('^(GENES_EXP|GENES_PU|CHEMCOMP|OTHER)?(,(GENES_EXP|GENES_PU|CHEMCOMP|OTHER))*$', AF_type):
      response['error'] = "subtype not correct"
      return response

   if not re.match ('^(allele|citation|homology|species|species_af|go_bioproc|go_molfunc|go_cellcomp|keggpw|keggeg)?(,(allele|citation|homology|species|species_af|go_bioproc|go_molfunc|go_cellcomp|keggpw|keggeg))*$', annotationNodes):
      response['error'] = "nodes not correct"
      return response

   if not re.match ('^(|d3-link|d3-charge|d3-center|ngraph|fmmm)$', layout):
      response['error'] = "layout not correct"
      return response


   AF_type         = AF_type.split (',')
   AF_species      = AF_species.split (',')
   annotationNodes = annotationNodes.split (',')

   if "" in AF_type: AF_type.remove ("")
   if "" in AF_species: AF_species.remove ("")

   # Search
   return search_utils.resolveAllAgeingFactors (response, AF_type, AF_species, annotationNodes, layout)



##---------------------------Start searchAgeingFactor------------------------------##
def searchAgeingFactor (request, response):
   """
   annotation nodes:
      gobp       DONE
      gocc       DONE
      gomf       DONE
      species    DONE
      homanob
      citation   DONE
      allele     DONE
      keggpw     DONE
      keggeg     DONE

   depth:
      true / false / <depth>

   layout:
      d3-link / d3-charge / d3-center / ngraph / fmmm /

   false: AF --> OB --> AF
   true:  AF --> OB --> AF --> OB --> AF

   AF --> GO(bio_proc. == true)
   AF --> SP
   """


   AF_id           = request.GET.get ("id")
   annotationNodes = request.GET.get ("nodes", "")
   depth           = request.GET.get ("depth", "false")
   layout          = request.GET.get ("layout", "")


   # Check input
   if len (AF_id) == 0:
      response['error'] = "no id defined"
      return response
   if not re.match ('^AF_[0-9]{6}(,AF_[0-9]{6})*$', AF_id):
      response['error'] = "id not correct"
      return response

   if not re.match ('^(allele|citation|homology|species|species_af|go_bioproc|go_molfunc|go_cellcomp|keggpw|keggeg)?(,(allele|citation|homology|species|species_af|go_bioproc|go_molfunc|go_cellcomp|keggpw|keggeg))*$', annotationNodes):
      response['error'] = "nodes not correct"
      return response

   if not re.match ('^(true|false|[0-9]+)$', depth):
      response['error'] = "depth not correct"
      return response

   if not re.match ('^(|d3-link|d3-charge|d3-center|ngraph|fr|fmmm)$', layout):
      response['error'] = "layout not correct"
      return response


   AF_id           = list (set (AF_id.split (',') ) )
   annotationNodes = annotationNodes.split (',')
   depth           = -1 if depth == "true" else 2 if depth == "false" else (int (depth) * 2)

   # Search
   return search_utils.resolveAgeingFactors (response, AF_id, annotationNodes, depth, layout)



##---------------------------Start searchAgeingFactorNeighbors---------------------##
def searchAgeingFactorNeighbors (request, response):
   AF_id = request.GET.get ("id")


   # Check input
   if len (AF_id) == 0:
      response['error'] = "no id defined"
      return response
   if not re.match ('^(GOI:)?AF_[0-9]{6}(,(GOI:)?AF_[0-9]{6})*$', AF_id):
      response['error'] = "id not correct"
      return response


   AF_id = list (set (AF_id.split (',') ) )

   # Search
   return search_utils.resolveAgeingFactorNeighbors (response, AF_id)



##---------------------------Start entry-------------------------------------------##
@csrf_exempt
def entry (request):
   """
   format:
      json / blub
   """

   response = {}

   resolveBodyParameters (request, response)

   type_     = request.GET.get ("type")
   format_   = request.GET.get ("format", "json")
   userData_ = request.GET.get ("userData")


   if userData_ != None: response['userData'] = userData_


   if "help" in request.GET:
      response["help"] = _help

   elif not type_:
      response['error'] = "No command defined; use api/search?help"

   elif type_ == "AllAgeingFactors" or type_ == 'AAF':
      searchAllAgeingFactors (request, response)

   elif type_ == "AgeingFactor" or type_ == 'AF':
      searchAgeingFactor (request, response)

   elif type_ == "AgeingFactorNeighbors" or type_ == 'AFN':
      searchAgeingFactorNeighbors (request, response)

   else:
      response['error'] = "Command not found"


   return formatResponse (response, format_)



##---------------------------Start resolveBodyParameters---------------------------##
def resolveBodyParameters (request, response):
   if len (request.body) == 0: return

   body = request.body.decode ('ascii')

   if request.GET.get ("body") == "true":
      GET  = request.GET.copy ()
      body = io.StringIO (body)

      for line in body:
         key, value = line.split ("=", 1)
         if key == "userData":
            GET["userData"] = body.read ()
            break
         if value[-1] == '\n': value = value[:-1]
         GET[key] = value

      request.GET = GET

   else:
      response['userData'] = body



##---------------------------Start formatResponse----------------------------------##
def formatResponse (response, format_):
   if format_ == "ogdf":
      if 'error' in response:
         return HttpResponse (response['error'], content_type="text/plain")

      result = "graph [\n"
      if 'nodes' in response:
         for node in response['nodes']:
            result += "node [ id " + node['id'] + " type " + node['type'] + " ]\n"
      if 'links' in response:
         for edge in response['links']:
            result += "edge [ source " + edge['source'] + " target " + edge['target'] + " ]\n"

      result += "]\n"

      return HttpResponse (result, content_type="text/plain")

   return JsonResponse (response)
