"""JANet REST API
"""

import os, sys
import io
import json
import re

from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt

BASE_DIR = os.path.dirname (os.path.dirname (os.path.abspath (__file__) ) )
sys.path.insert (0, os.path.join (BASE_DIR, "Tools") )

from Arguments import Arguments
from Neo4j import Neo4j


neo4j = Neo4j ()



##---------------------------Help Message------------------------------------------##
_help = """\
TODO: write this help page ;p
"""



##---------------------------Start startup-----------------------------------------##
def startup (request, response):
   response["species_gE"] = species_gE = []
   response["species_gP"] = species_gP = []
   response["species_cp"] = species_cp = []
   response["species_of"] = species_of = []
   response["gE"]         = gE         = []
   response["gP"]         = gP         = []
   response["chemcomp"]   = chemcomp   = []
   response["other"]      = other      = []


   ##------------------------Get all putative gene AFs-----------------------------##
   for record in neo4j.call ('get/AF-gene-putative', {}):
      gP.append (dict (record.items () ) )

   for record in neo4j.call ('get/AF-gene-putative-species', {}):
      species_gP.append (dict (record.items () ) )

   ##------------------------Get all non-putative gene AFs-------------------------##
   for record in neo4j.call ('get/AF-gene-non_putative', {}):
      gE.append (dict (record.items () ) )

   for record in neo4j.call ('get/AF-gene-non_putative-species', {}):
      species_gE.append (dict (record.items () ) )

   ##------------------------Get all compound AFs----------------------------------##
   for record in neo4j.call ('get/AF-compound', {}):
      chemcomp.append (dict (record.items () ) )

   for record in neo4j.call ('get/AF-compound-species', {}):
      species_cp.append (dict (record.items () ) )

   ##------------------------Get all other_factor AFs------------------------------##
   for record in neo4j.call ('get/AF-other_factor', {}):
      other.append (dict (record.items () ) )

   for record in neo4j.call ('get/AF-other_factor-species', {}):
      species_of.append (dict (record.items () ) )


   return response



##---------------------------Start layout------------------------------------------##
def layout (request, response):
   if request.method != 'POST':
      response['error'] = "Request must be a HTTP POST"
      return response

   layout = request.GET.get ("layout", "")

   # Check input
   if len (request.body) == 0:
      response['error'] = "missing list of links"
      return response

   if not re.match ('^(fmmm)$', layout):
      response['error'] = "layout not correct"
      return response

   content = request.body.decode ('ascii')

   if not re.match ('^[a-zA-Z0-9\'".\\n\t:\{\[\]\} ,_-]+$', content):
      response['error'] = "list of links is not correct"
      return response

   try:
      links = json.loads (content)
      # print (content)
   except:
      response['error'] = "cannot parse list of links, it must be a JSON file"
      return response

   nodes_ = []
   try:
      for link in links:
         nodes_.append (link['source'])
         nodes_.append (link['target'])
      nodes = []
      for id in list (set (nodes_) ):
         nodes.append ({'id': id})

   except:
      print (link)
      raise


   response['config'] = {'warmup': 0, 'cooldown': 0}
   import GraphLayout
   GraphLayout.computeLayout (nodes, links, "FMMM")

   response['nodes'] = nodes_= {}
   for node in nodes:
      nodes_[node['id'] ] = {'x': node['x'], 'y': node['y'], 'z': node['z']}

   return response



##---------------------------Start entry-------------------------------------------##
@csrf_exempt
def entry (request):
   response = {}

   type_ = request.GET.get ("type")

   if "help" in request.GET:
      response["help"] = _help

   elif not type_:
      response['error'] = "No command defined; use api/special?help"

   elif type_ == "startup":
      startup (request, response)

   elif type_ == "layout":
      layout (request, response)

   else:
      response['error'] = "Command not found"

   return JsonResponse (response)
